/*++++++++++++++
 Copyright:    (C) 2008-2015 UDS/CNRS
 License:       GNU General Public License
.IDENTIFICATION parfile.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Bib datafiles
.KEYWORDS       Index based on Paragraphs
.VERSION  1.0   25-Jun-1991
.VERSION  1.1   05-Jul-1991: No blank line in Display
.VERSION  1.2   04-Mar-1993: -b & -d Options
.VERSION  1.3   17-Mar-1993: Can have several input files.
		Lines starting with # are comments and just output
.VERSION  1.4   29-Mar-1993: Store index in NET format
.VERSION  1.5   03-Apr-1993: Fixed problems in Merge
.VERSION  1.6   09-Apr-1993: Fixed problem when looking at end of file
.VERSION  1.7   20-Apr-1993: Bug in Line Index ...
.VERSION  1.8   11-Apr-1994: Pb in Display with several files
.VERSION  2.0   17-May-1994: Added General Index
.VERSION  2.1   09-Sep-1994: Expandable buffer for stdinput
.VERSION  2.2   12-Dec-1994: Options  -f (Dictionary) +  -n (Display byte)
.VERSION  2.3   03-Mar-1995: -T option (temporary directory)
.VERSION  2.4   30-Mar-1995: -n option works also with line numbers
.VERSION  2.5   19-Apr-1995: Transformation of plurals modified
.VERSION  2.6   06-Jul-1995: No message of transformation for truncated words
.VERSION  2.7   20-Jul-1995: Keep words starting by numbers but having alpha
.VERSION  2.8   26-Mar-1996: Keep 2-letter words in all uppercase
.VERSION  2.9   16-Apr-1996: -gs option.
.VERSION  3.0   17-Apr-1996: Store the delimiter with indexes.
.VERSION  3.1   09-Jan-1997: Silent also for disgarded words...
.VERSION  3.2   06-Aug-1997: Rule 'ies' into 'y', allow ' in words...
.VERSION  3.3   26-Aug-1997: There was a problem with "-gs" opttion,
			dictionary loaded only AFTER substitution!!!
			Fixed.
.VERSION  3.4   22-Oct-1997: There was a problem with fully uppercase...
			Fixed.
.VERSION  3.5   28-Jan-1998: -merge option revisited --- merging stars ONLY
			from line number specified.
.VERSION  3.6   13-Feb-1998: option -tex, and dichotomic search in LINDEX.
.VERSION  3.7   03-Jul-1998: option -tex: convert _.._  ^..^
.VERSION  3.8   15-Oct-1998: option -inum (index numeric)
.VERSION  3.9   21-Oct-1998: KEEP the words specified in equivalences...
.VERSION  4.0   22-Jan-1999: Sort INSIDE the program.
.VERSION  4.1   27-Feb-1999: -lw option
.VERSION  4.2   16-May-1999: -itex option
.VERSION  4.3   21-Jul-1999: \i was not defined...
.VERSION  4.4   28-Feb-2000: -e option
.VERSION  4.5   07-Jun-2000: Accept -e- (negative regular expressions)
.VERSION  4.6   09-Jul-2000: Syndrome: Ja{\'e}n --> Jan --> removed !
.VERSION  4.7   20-Apr-2001: -n option: use 2 numbers in case of multiple files
			When a word is modified by a rule, it's never
			removed from the index.
.VERSION  4.8   09-Jun-2001: Create index from specified lines (-x option)
.VERSION  4.9   14-Jan-2002: -e executed also on stdin.
.VERSION  4.91  12-May-2002: Option -gs there was a bug in the replacement
.VERSION  5.0   17-Mar-2003: Option -gb to take keys by best matching order
				+ option -w
.VERSION  5.1   12-Apr-2005: Cosmetics -- install as several programs
			parfile  parindex   pargrep  paregrep
.VERSION  5.2   05-May-2006: -ii kept in index
.VERSION  5.21  04-Oct-2006: Bug in stumatch ??
.VERSION  5.22  08-Nov-2006: Insert the question in the answer (%%)
.VERSION  5.23  14-Nov-2006: Insert any comment following the numbers
.VERSION  5.3   18-Nov-2006: pargrep, parsort
.VERSION  5.4   22-Dec-2006: merge in parsort (-m option)
.VERSION  5.41  30-May-2007: Bug in readPar
.VERSION  5.42  04-Jun-2007: No title in simple parfile
.VERSION  5.43  27-Oct-2007: Compatible with grep for -w option (\< and \>)
.VERSION  5.44  21-Mar-2008: pb in sort
.VERSION  5.45  12-Apr-2008: parsort bug (-r / -n inverted)
.VERSION  5.46  29-Jul-2008: Accept line numbers fir 1st paragraph
.VERSION  5.47  26-Jul-2009: -p option
		Option -p0 (no intro at all) / -p1 (one line)
.VERSION  5.48  30-Aug-2009: -k  in parsort: accept a delimiter
.VERSION  5.49  20-Nov-2009: Don't index first paragraph
.VERSION  5.50  21-oct-2013: do not use plural when word is in uppercase
.VERSION  5.51  21-jan-2014: 
.VERSION  5.52  13-Mar-2014: add hostname
.VERSION  5.53  14-Jun-2014: Cosmetics (use SEARCH_SILENT)
.VERSION  5.54  13-Aug-2014: bug with SEARCH_SILENT
.COMMENTS
	This program deals with files organized in paragraphs, i.e.
	a record is splitted into paragraphs. Paragraphs are separated
	by a single blank line.
	Such a file has normally a line index with an extension .i0 .
		made of 8 bytes: linenumber of blank line + offset
	and a General Index (words) with an extension specified by -x (.i-)
---------------*/

#define VERSION "5.54"


#define MAXWORDLEN	24	/* Words larger than MAXWORDLEN are truncated */
static int maxwordlen = 7 ;	/* Default is to keep obnly 7 characters/word */
#define MAXWORDS	24	/* Max number of words in a Query: 1 byte
                                   for count, 3 bytes for a mask 	*/
#define LENLINE1	256	/* Length of First Line			*/

/* Sorting Parameters 	*/
#ifndef SBUF_SIZE       /* Blocks to save strings */
#define SBUF_SIZE       (16<<10) /* 16k */
#endif
#ifndef SDATA_SIZE      /* Blocks to save wordo's */
#define SDATA_SIZE      1024
#endif
#ifndef SPART_SIZE
#define SPART_SIZE      64
#endif

/*#define isaword(c)	(isalnum(c)||((c)=='\''))*/
#define isaword(c)	isalnum(c)
#define isawtex(c)	(isaword(c)||(*p == '\\')||(*p == '{')) /*}*/

static char *xtension = ".i-" ;
static char *echo ;		/* If echo the questions */
static int addIntro = 2;	/* Add an introduction + blank line */
static char *progname;

/* Different kinds of grep */
#define GREPdelimited	010
#define GREP		000
#define EGREP		001
#define FGREP		002
#define EGREPdelimited	011
/*
#define PARFILE 	0
#define PARGREP		1
#define PAREGREP	2
#define PARFGREP	3
#define PARINDEX	4
#define PARSORT		5
#define PARMERGE	6
*/

static char usage[] = "\
Usage: parfile [-dexpr] [-echo] [-b|-g[sb#]|-e /regexpr/] [-f dic] [-i2]\n\
       [-index|-i[i]-[%...]|-i/regexp/] [-inum] [-itex] [-lw wordsize] \n\
       [-n[text]] [-w[text] [-sb bsize] [-texs] [-v[pc]] [-x xtension] \n\
       [-p] parfile... [-merge|-replace infile]";
static char help[] = "\
   -dexpr: Define delimiter expression (anchored at line beginning)\n\
           -d alone stops at end-of-line; default is paragraph\n\
           (may contain C-like controls e.g. \\n)\n\
    -echo: Copy to output the question, starting by a #; echoes also comments\n\
       -b: Input consists of Byte position (faster!)\n\
  -e /re/: Locate paragraphs satisfying the regular expression, \n\
	   or NOT statisfying when the / is preceded by - ! or ^\n\
	   (executed on stdin when no parfile specified on command line)\n\
   -f dic: Dictionary (#Keep:  and #Change:) specified\n\
       -g: Input consists of Keys to look at\n\
      -gb: as -g, keys are anded and sorted (best matching first)\n\
	   (may include a number = how many scores to list, default=1)\n\
      -gs: as -g, but with silent replacement\n\
      -i2: Keep also the words made of two letters (-i1: keep 1-letter words)\n\
    -inum: index also the numbers (numbers are not indexed by default)\n\
    -itex: index converts words from LaTeX (no macro substitution!)\n\
      -i-: Create the GeneralIndex (extension .i-); -ii- folds to lowercase\n\
           The %-specs tell which titles to keep (e.g. -i-%AT)\n\
                      or to remove (e.g. -i-%-J  to remove %J)\n\
  -i/reg/: Create the index from a Regular Expression \n\
 -i^/reg/: Don't take lines matching reg in the index\n\
   -index: Create the LineIndex file (extension .i0)\n\
  -lw len: Specifies the maximal size of words (default: 7)\n\
 -n[text]: Write out the Byte Number preceded by text, e.g.  -n%#\n\
   -merge: infile merged with paragraphs extracted from parfile:\n\
           (input is where_number, paragraph_to_insert)\n\
       -p: no header in the output (plain)\n\
 -replace: similar to -merge, but the line is replaced\n\
 -sb size: size of blocks for sorting (default 1024)\n\
    -texs: convert special symbols like ~(about) <~ (la) into Tex\n\
    -x xt: specified the extension (default .i-) for the index\n\
    -v...: verbose mode, flags words more frequent than the percentage\n\
    -wtxt: Write out the Weight of the result, followed by matching words\n\
  parfile: file(s) made of paragraphs\n\
The output consists of paragraphs from parfile, preceded by a 2-lines header;\n\
the header may be removed with the -p option.\n\
" ;

/* LaTeX Equivalences */
static char tex_symbols[] = "\
\nAA=A\nAE=AE\nL=L\nO=O\nOE=OE\
\naa=a\nae=ae\nc=\ni=i\nl=l\nlbrack=[\no=o\nrbrack=]\nss=ss\
\nv=\
\n";

/* \nastrono\nastroph\naug\ */
static char wchange[] = /* Words to remove ORDERED ! */ "\
\nadd\nadded\nalso\nand\napr\nare\
\naug\
\nbeen\nbeing\ncan\ncatalog=catalogue\ncentre=center\ncolour=color\ncom\
\ndec\nedu\nfeb\nfor\nfrom\ngov\
\nhas\nhave\nhaving\nhttp\niii\ninfrared=ir\ninto\
\njan\njul\njun\njune\
\nmade\nmake\nmar\nmay\nmil\nneb=nebula\nnet\nnot\nnov\noct\norg\
\npn=planeta nebula\npsr=pulsar\nqso=quasar\nqsos=quasar\
\nsep\nsn=supernova\nspectru=spectra\nsupplem=suppl\
\nthan\nthat\nthe\nthem\nthere\nthese\nthey\nthis\nultraviolet=uv\nvan\
\nvii\nvitesse=velocit\nvon\nwas\nwere\nwhen\nwhich\nwith\nwww\
\n";
static char *word_change = wchange;
static int lword_change = sizeof(wchange) - 1;

static char wkeep[] = /* Words for which we don't remove the ending 's'
			  or we keep the 'umlaut' sign
			  or 2 letters or less that we keep		*/ "\
\nir\njeans\nlens\nparkes\
\nphotoef\nphotoel\nphotoem\npks\npn\
\nsn\nstokes\nuv\nx\
\n";
static char *word_keep = wkeep;
static int lword_keep = sizeof(wkeep) - 1;
static char *dic_file = (char *)0;
static char *dic_cont = (char *)0;

#define  Perror		perror("*** parfile")
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#ifndef O_BINARY
#define O_BINARY 0
#endif

		/* In MSDOS, lines are terminated by \r\n ... */
#ifdef __MSDOS__
# ifndef INT4
# define INT4		long
# endif
# define BUFSIZE	4096
# define MAXFILES	32	/* Maximal Number of Files */
# define MAXOPENED	12	/* Maximal Number of Opened Files */
# define htonl(x)	(x)	/* No byte swap	*/
# define ntohl(x)	(x)	/* No byte swap	*/
# define gethostid()	0
# define getpid()	1
# define strncasecmp	strncmpi
# define strcasecmp	strcmpi
# define EOL		'\r'
  static char *delim="\r\n";		/* Delimiter expr. */
#else
# ifndef INT4
# define INT4		int
# endif
# include <unistd.h>		/* lseek, etc...	   */
# include <netinet/in.h>	/* hton (Host to Net)	   */
# include <sys/stat.h>		/* File status stat(2)	   */
# define BUFSIZE	8192
# define MAXFILES	64	/* Maximal Number of Files */
# define gethostid()	time(0)
# define EOL		'\n'
# define MAXOPENED	 4	/* Maximal Number of Opened Files */
  static char *delim="\n";		/* Delimiter expr. */
#endif
#define MAX_LOC 	0x7fffffff	/* Highest INT4 value 	*/
static char *Tdir = "" ;		/* Sorting Directory 	*/
static int nnl = 0;			/* Number of Newlines in delim */
static int lreg = 0;			/* Length of reg.exp	*/

static int articles ;

#define bcopy(s,d,l)	memcpy(d,s,l)
#define ITEMS(s)	(sizeof(s)/sizeof(s[0]))
#define MIN(a,b)	( a <= b ? a : b)
#define MAX(a,b)	( a >= b ? a : b)

typedef int (*SORT_FCT)(/* const void *, const void * */) ;

/* Merging files */
typedef struct {
    FILE *file;
    char *par;
    int bytes;		/* Allocated bytes for par */
} MERGEFILE;


/* The sorting method used here stores the data (WORDO entities)
   in data blocks, which are pointed to by partition blocks.
   When a block is full, its partition father adds another block.
*/

typedef struct { char *word ; long offset; } WORDO ;
typedef struct s_buf { 		/* To save strings (text) 	*/
    char *signature ;
    int allocated, used ;	/* Number of bytes used 	*/
    struct s_buf *previous ;	/* These buffers are chained	*/
    char buf[1] ;		/* Actual size in allocated 	*/
} SBUF;
typedef struct s_data {		/* To save WORDO entities	*/
    char *signature ;
    int allocated, used ;	/* Number of bytes used 	*/
    WORDO low, up ;		/* Limits of string [ .. .[	*/
    struct s_part *apart ;	/* Parent structure		*/
    WORDO record[1] ;		/* Actual size in allocated 	*/
} SDATA;
typedef struct s_part {		/* To save _data entities	*/
    char *signature ;
    int allocated, used ;	/* Number of bytes used 	*/
    WORDO low, up ;		/* Limits of string [ .. .[	*/
    struct s_part *apart ;	/* Parent partition, if any	*/
    struct s_data *adata[1] ;	/* Actual size in allocated 	*/
} SPART;

static char *sign_data = "@+o" ;
static char *sign_part = "par" ;
static char *sign_buf  = "buf" ;
#if 0
static WORDO low_word = { "", 0 } ;
static WORDO up_word  = {
    "\177\177\177\177\177\177\177\177\177\177\177" ,
    0x7fffffff } ;
#endif
static int sdata_size = SDATA_SIZE ;
static int spart_size = SPART_SIZE ;
static SBUF  *sbuf1 ;
static SPART *spart ;

typedef struct {	/* Index Line ==> Byte */
	INT4 lineno;	/* --- Stored in Net format --- */
	INT4 offset;	/* --- Stored in Net format --- */
    } LINDEX;
typedef struct {
  	char *name;
	char *delim;	/* Delimiter for this file	*/
	char *idb;	/* Loaded line index		*/
  	LINDEX *idx, *ide;
	char *gindex, *agindex;	/* General Index	*/
	int  lgindex, lagindex; /* Allocated lengths	*/
	int           bagindex; /* Used      lengths	*/
  	int   db;			/* File Number 	*/
  	int   gf;	/* General Index File Number 	*/
  	int   rank;	/* To know which file to close 	*/
  } PARFILE;
static PARFILE parfile[MAXFILES];
static int parfiles;	/* Number of Parfiles */
static int opened;	/* Number of Opened Parfiles */
static int orank;	/* Increased at each open */
static int multiple;	/* 1 if several files */

static char buf[BUFSIZE+1];
static char *word_abuf, *word_ebuf ;	/* Shared by get_word routines	*/
static int (*word_fct)() ;		/* Shared by get_word routines  */
static char btex[60] ;	/* Buffer used in -tex option, shared by pr_tex	*/
static char otex= 0;	/* -texs   Option  */
static char ov  = 0;	/* Verbose Option  */
static char oc  = 0;	/* Create option   */
static char oc1 = 0;	/* First byte	   */
static char og  = 0;	/* -g option   	   */
#define SEARCH_SILENT 2	/* -gs option	   */
#define SEARCH_BEST 4	/* -gb option	   */
static int oglist = 1;	/* -gb # items     */
static char orm = 0;	/* remove option (-replace)   */
static char opti2 = 0;	/* -i2 option      */
static char optii = 0;	/* -ii option      */
static char obyte = 0;	/* Option byte off */
static char squeeze = 1;/* Just once       */
static char inMerge;    /* Set by Merge    */
static char *mfile;	/* Merge File Name */
static char *ogk;	/* List of %to keep in general index (-i- option) */
static char *o_n;	/* -n option	   */
static char *o_w;	/* -w option	   */
static int thefileno ;	/* with -n option  */
static char inumop;	/* -inum option    */
static char itexop;	/* -itex option    */

typedef struct link1 {	/* Single-linked list 		*/
	struct link1 *previous;
	char *text;
	char neg ;	/* NEGATIVE (0x01) regexp	*/
	char circf ;	/* ^ in regular expression	*/
	char word;	/* -w option (<word> grep)	*/
	char sortsep;	/* Separator for sort (-k)	*/
	char buf[1]; 	/* Actual size depends on alloc */
} LINK1 ;
static LINK1 *grep ;	/* -e option	   */
static LINK1 *reg_keep = (LINK1 *)0;	/* -i/reg_exp/	   */
static LINK1 *reg_leave= (LINK1 *)0;	/* -i!/reg_exp/	   */
static LINK1 *sort_keys= (LINK1 *)0;	/* -k sort options */
static int nfgrep = 0;	/* Number of files */

/*==================================================================*/
static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for (s = text; *s && (*s != c); s++) ;
    return(s-text);
}

static int stuloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character, case insensitive
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s, uc;
    uc = tolower(c);
    for (s = text; *s && (tolower(*s) != uc); s++) ;
    return(s-text);
}

static int stucmp(s1, s2)
/*++++++++++++++++
.PURPOSE  Compare (case insensitive) 2 strings
.RETURNS  Difference s1 - s2
-----------------*/
  char *s1;     /* IN: First string */
  char *s2;	/* IN: Secnd string */
{
  int dif;
    for (dif=0; (dif == 0) && *s1; s1++, s2++)
	dif = tolower(*s1) - tolower(*s2);
    if (dif == 0) dif = tolower(*s1) - tolower(*s2);
    return(dif);
}

static char *stustr(text, str)
/*++++++++++++++++
.PURPOSE  Locate substring "str" within string "text"
.RETURNS  Pointer in text / NULL
-----------------*/
  char *text;     /* IN: String to interpret */
  char *str;	  /* IN: Substring to locate */
{
  int i, dif, uc;
    uc = tolower(*str);
    while (1) {
        while (*text && (tolower(*text) != uc)) text++;
	if (!*text) return((char *)0);
	for (dif=0, i=1; (dif==0) && str[i]; i++)
	    dif = tolower(text[i]) - tolower(str[i]);
	if ((dif==0) && (!str[i]))
	    return(text);
        text++;
    }
    return((char *)0);
}

static char *parloc(text, str)
/*++++++++++++++++
.PURPOSE  Locate specified string in col.1
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  char *str; 	  /* IN: Substring to locate */
{
  int i;
    while (*text) {
        for (i=0; str[i] && (str[i] == text[i]); i++) ;
	if (str[i] == 0) return(text);
	text += strloc(text, '\n');
	if (*text) text++;
    }
    return((char *)0);
}

#ifdef NO_strdup
static char *strdup(text)
/*++++++++++++++++
.PURPOSE  Duplicate (clone) a string
.RETURNS  New string
-----------------*/
  char *text;     /* IN: String to duplicate */
{
  char *s;
    if (s = malloc(1+strlen(text))) strcpy(s, text);
    return(s);
}
#endif

static char *adate(t)
/*++++++++++++++++
.PURPOSE  Edition of date/time
.RETURNS  Static with edited date/time
-----------------*/
  time_t t;	/* IN: Date/Time */
{
static char ctime_order[] = { 9, 10, 11, 5, 6, 7, 8, 21, 22, 23, 24, 20,
    12, 13, 14, 15, 16, 17, 18, 19 };
static char st_date[sizeof(ctime_order)+1];
  char *s; int i;
    s = ctime(&t);

    for(i=0; i<sizeof(ctime_order); i++) {
	st_date[i] = s[ctime_order[i]-1];
	if ((i<1) || (i>sizeof(ctime_order)-9)) st_date[i] |= '0';
    }
    return(st_date);
}

static INT4 a6tol(buf)
/*++++++++++++++++
.PURPOSE  Interpret a set of characters into a 32-bit (64-base)
.RETURNS  Equivalent number
.REMARKS  End-of-line is the highest number 0x7fffffff
-----------------*/
  char *buf;	/* IN: ascii representation  */
{
  INT4 r;
    if (*buf == '\n') r = MAX_LOC;
    else for (r=0; isgraph(*buf); buf++)  r = (r<<6) + (*buf-'0');
    return(r);
}

static char *ltoa6(val)
/*++++++++++++++++
.PURPOSE  Convert to ascii (64-base) a positive INT4
.RETURNS  Equivalent number
.REMARKS  End-of-line is the highest number 0x7fffffff
-----------------*/
  INT4 val;	/* IN: value to translate to ascii */
{
  static char buf[12];
  char *p ;
    p=buf+sizeof(buf)-1;
    if (val == MAX_LOC) *--p = EOL;
    else { do *--p = '0'+(val&077), val>>=6 ; while (val); }
    return(p);
}

static char *load_stream(f)
/*++++++++++++++++
.PURPOSE  Load the contents of stdin (or other stream)
.RETURNS  The loaded text terminated by NULL / NULL if error
-----------------*/
  int f;	/* IN: Strema fileno */
{
  char *buffer, *b, *e;
  long bytes = (32<<10)-1;	/* 32K */
  int  bexp  =  64<<10;		/* Expand size */
  int n;

    buffer = b = malloc(bytes+1);
    e = b + bytes;
    while ((n = read(f, b, e-b)) > 0) {
	b += n;
	if ((e-b) >= 128) continue;
	/* Must expand */
	n = b - buffer;	/* Current length */
	bytes += bexp;
	buffer = realloc(buffer, bytes+1);
	b = buffer + n;
	e = buffer + bytes;
	if (bexp < (4<<20))	/* Expand size until 4Mbytes */
	    bexp += 64<<10;
    }
    if (n<0) {
	fprintf(stderr, "****file#%d", f);
	perror("");
	return((char *)0);
    }
    *b = 0;
    return(buffer);
}

static char *load_file(name, obuf, show)
/*++++++++++++++++
.PURPOSE  Load a file completely into memory
.RETURNS  The loaded text terminated by NULL / NULL if error
.REMARKS  name="-" ==> stdin
	Skip first paragraph when obuf != NULL
-----------------*/
  char *name;	/* IN: Name of the file to load */
  char *obuf;	/* IN: Buffer (expanded)	*/
  int show;	/* IN: Non-zero to show errors	*/
{
  int  n, f;
  long fsize;
  char *buffer, *b;

    /* Stdin: append */
    if ((name[0] == '-') && (!name[1])) {
	b = load_stream(0);
	if (!obuf) return(b);
	n = strlen(obuf);
	buffer = realloc(obuf, n+strlen(b)+1);
	strcpy(buffer+n, b);
	return(buffer);
    }

    /* Load file into Memory */
    f = open(name, O_BINARY);
    if (f <= 0) {
    	if (show) { fprintf(stderr, "****load_file(%s)", name); perror(""); }
    	return ((char *)0);
    }
    fsize = lseek(f, 0L, 2); lseek(f, 0L, 0);
    n = 0;			/* How many bytes in buffer */
    if (obuf) {
	n = strlen(obuf);
	buffer = realloc(obuf, fsize+n+1);
    }
    else buffer = malloc(fsize+1);

    /* Verify memory was allocated */
    if (!buffer) {
	fprintf(stderr, "****load_file(%s)", name);
	perror("");
	return(buffer);
    }

    /* Copy the file into memory */
    b = buffer + n;		/* Where to add the new file */

    if (read(f, b, fsize) != fsize) {
	fprintf(stderr, "****load_file(%s[%ldbytes])", name, fsize);
    	perror("");
    	free(buffer);
    	return ((char *)0);
    }
    b[fsize] = 0;
    close(f);

    /* Remove first paragraph */
    if (obuf && delim[0]) {
      int i; char *p;
        p = b;
        while (*p) {
	    for (i=0; delim[i] && (p[i] == delim[i]); i++) ;
	    if (delim[i] == 0) break;	/* End of Paragraph found */
	    p += i;
	    p += strloc(p, '\n');
	    if (*p) p++;
	}
	if (p != b) 			/* Omit the first paragraph */
	    strcpy(b, p);
    }

    return(buffer);
}

/*==================================================================*/
static int Usage()
/*++++++++++++++++
.PURPOSE  Just print the "Usage"
.RETURNS  -1
-----------------*/
{
   puts(usage);
   return(-1);
}

static int Help()
/*++++++++++++++++
.PURPOSE  Just print a help summary
.RETURNS  -1
-----------------*/
{
    Usage();
    printf("%s", help) ;
    return(-1);
}

static int check_range(value, minval, maxval, text)
/*++++++++++++++++
.PURPOSE  Set a value in range, issue a message if outside
.RETURNS  The substituted value
-----------------*/
  int value ;	/* IN: Value to check */
  int minval ;	/* IN: Lower bound    */
  int maxval ;	/* IN: Upper bound    */
  char *text ;	/* IN: Text */
{
    if (value < minval) {
	fprintf(stderr, "++++ Parameter %s too small (%d), set to %d\n", text,
	    value, minval) ;
	return(minval) ;
    }
    if (value > maxval) {
	fprintf(stderr, "++++ Parameter %s too large (%d), set to %d\n", text,
	    value, maxval) ;
	return(maxval) ;
    }
    return(value) ;
}

static char *Gets()
/*++++++++++++++++
.PURPOSE  Get a line (and forget the end of) on stdin
.RETURNS  as gets
-----------------*/
{
  static char *abuf; static int lbuf;
  int i;

    if (!lbuf) { lbuf = BUFSIZ-1; abuf = malloc(lbuf+1); abuf[lbuf] = 0; }

    if (! fgets(abuf, lbuf-1, stdin)) {	/* Error or EOF ? */
	if (ferror(stdin)) perror("****Gets");
	return((char *)0);
    }
    i = strloc(abuf, '\n');
    while (!abuf[i]) {
	lbuf += BUFSIZ; abuf = realloc(abuf, lbuf);
	if (!fgets(abuf+i, BUFSIZ, stdin)) break;
	i = strloc(abuf, '\n');
    }
    abuf[i] = 0;
    /* if (echo) printf("%s%s\n", echo, abuf) ; */
    return(abuf);
}

static char *readPar(mf)
/*++++++++++++++++
.PURPOSE  Get a Paragraph (defined by delim)
.RETURNS  Buffer with result / NULL
.REMARKS  The output does NOT includes the delimiter.
-----------------*/
  MERGEFILE *mf ; 	/* IN: Paragraph file to read */
{
  char *abuf, *rbuf, *ebuf ;
  int n, ldel, bol=1;

    if (!mf->par)
	mf->par = malloc(mf->bytes = BUFSIZ);
    ldel = strlen(delim);
    abuf = mf->par;
    ebuf = abuf + mf->bytes;
    rbuf = ebuf - 2 - ldel;	/* Leave room to read the delimiter */

    while (1) {
	if (abuf >= rbuf) {
	    n = abuf - mf->par;
	    mf->bytes += BUFSIZ;
	    mf->par = realloc(mf->par, mf->bytes);	/* V5.41 */
	    abuf = mf->par + n;
	    ebuf = mf->par + mf->bytes;
	    rbuf = ebuf - 2 - ldel;
	}
        *abuf = 0;
	if (! fgets(abuf, ebuf-abuf, mf->file)) 	/* End-Of-File */
	    break;
	if (bol && (strcmp(abuf, delim) == 0)){	/* Delimiter Found.... */
	    *abuf = 0;
	    break;
	}
	abuf += strloc(abuf, '\n') ;
	if ((bol = *abuf)) abuf++ ;
    }

    abuf += strlen(abuf);
    if (abuf == mf->par) {			/* Nothing read: EOF? */
	if (feof(mf->file))
	    return((char *)0);
    }

    /* Skip the empty lines */
    while (1) {
    	bol = fgetc(mf->file);
	if (bol == EOF) break;
	if (bol == '\n') 	/* Empty line */
	    continue;
	ungetc(bol, mf->file);
	break;
    }

    return(mf->par);
}

static char *getPar(f)
/*++++++++++++++++
.PURPOSE  Get a Paragraph (defined by delim)
.RETURNS  The paragraph / NULL when file exhausted
-----------------*/
  FILE *f ; 	/* IN: Input file to read */
#if 0
{
  static char *abuf, *ebuf, *rbuf;
  char *b ;
  int i, len, bol, eof, ldel;

    ldel = strlen(delim) ;
    bol  = 1 ;
    eof  = 0 ;
    b = abuf ;

    while (1) {
	if (b >= rbuf) {
	    i = b - abuf ;
	    if (!abuf) i = 0, abuf = malloc(len = 4*BUFSIZ) ;
	    else abuf = realloc(abuf, len = (ebuf-abuf) + 4*BUFSIZ) ;
	    ebuf = abuf + len ;
	    rbuf = ebuf - 2 - ldel ;
	    b = abuf + i ;
	}
	if (! fgets(b, ebuf-b, f)) { eof = 1; break ; }	/* End of File */
	if (bol && (strcmp(b, delim) == 0)) { 	/* Delimiter Found.... */
	    if (b == abuf) continue ;
	    *b = 0 ;
	    break ;
	}
	b += strloc(b, '\n') ;
	if (bol = *b) b++ ;
    }

    if (eof && (b == abuf)) return((char *)0) ; 	/* End of File */
    return(abuf);
}
#else		/* use readPar */
{
  static MERGEFILE mf;
    mf.file = f;
    return(readPar(&mf));
}
#endif

/*==================================================================
		Sorting Facility
 *==================================================================*/

static char *add_word(s)
/*++++++++++++++++
.PURPOSE  Add a string into SBUF
.RETURNS  The copy of the string
-----------------*/
  char *s ;	/* IN: String to add */
{
  int len, bfree ;
  SBUF *b ; char *p ;
    len = strlen(s) + 1 ;
    bfree = sbuf1 ? sbuf1->allocated - sbuf1->used : 0 ;
    if (len > bfree) {		/* Need another block ...	*/
        b = (SBUF *)malloc(sizeof(SBUF) + SBUF_SIZE - 1) ;
        b->signature = sign_buf ;
        b->previous  = sbuf1 ;
        b->allocated = SBUF_SIZE ;
        b->used = 0 ;
        sbuf1 = b ;
    }
    p = sbuf1->buf + sbuf1->used ;
    strcpy(p, s) ;
    sbuf1->used += len ;
    return(p) ;
}

static SDATA *new_data(n, low, up)
/*++++++++++++++++
.PURPOSE  Create a new data block of n records.
.RETURNS  The new data block. NO Parent !
-----------------*/
  int n ; 	/* IN; Numbre of records */
  WORDO *low ;	/* IN: lower bound */
  WORDO *up  ;	/* IN: upper bound */
{
  SDATA *P ;
    P = (SDATA *)malloc(sizeof(SDATA) + (n-1)*sizeof(WORDO)) ;
    memset(P, 0, sizeof(SDATA)) ;
    P->signature = sign_data ;
    P->allocated = n ;
    P->low = *low ; P->up = *up ;
    if (strcmp(P->low.word, P->up.word) >= 0) {
	fprintf(stderr, "****Problem new_data(%d, %s/%ld, %s/%ld)\n",
	n, P->low.word, P->low.offset, P->up.word, P->up.offset) ;
    }
    return(P) ;
}

static SPART *new_part(n, low, up)
/*++++++++++++++++
.PURPOSE  Create a new partition of n elements
.RETURNS  The new partition
-----------------*/
  int n ; 	/* IN; Numbre of records */
  WORDO *low ;	/* IN: lower bound */
  WORDO *up  ;	/* IN: upper bound */
{
  SPART *P ;
    P = (SPART *)malloc(sizeof(SPART) + (n-1)*sizeof(SDATA *)) ;
    memset(P, 0, sizeof(SPART)) ;
    P->signature = sign_part ;
    P->allocated = n ;
    P->low = *low ; P->up = *up ;
    return(P) ;
}

static void expand_data(block)
/*++++++++++++++++
.PURPOSE  Add a new record in a data block.
.RETURNS  ---
-----------------*/
  SDATA *block ;
{
  SPART *P, *P1 ;
  SDATA **pd, **pe, *nb ;
  WORDO *w0, *w1, *w2, *wc ;
  static WORDO wordo ;

    P = block->apart ;		/* Partition to which the block belongs */
    for (pd = P->adata, pe = pd + P->used ; (pd < pe) && (*pd != block); pd++) ;
    if (*pd != block) {
        fprintf(stderr, "****Basic Problem in expand_data(%p)\n", block) ;
        exit(1) ;
    }

    if (P->used >= P->allocated) {	/* A new partition is required... */
        P1 = new_part(P->allocated, &(block->low), &(block->up)) ;
	P1->apart = P ;
        *pd = (SDATA *)P1 ;
	block->apart = P1 ;
        P = P1 ;
        P->used = 1 ;
        pd = P->adata ;
        *pd = block ;
    }
#ifdef DEBUG
    if (strcmp(block->low.word, block->up.word) >= 0) {
        fprintf(stderr, "****Problem expand_data(%s/%d, %s/%d)\n",
        block->low.word, block->low.offset, block->up.word, block->up.offset) ;
    }
#endif

    /* The block has to be split. Decide the middle limit */
    w0 = block->record ;
    w1 = block->record + block->used ;
    w2 = block->record + block->used/2 ;
    /* Try to keep the same word in a block, by moving to left... */
    wc = w2 ;
    wordo.offset = 0 ;
    for (wc = w2-1; (wc->word == w2->word) && (wc >= w0); wc--) ;
    ++wc ;
    if (wc == w0) { /* ... and moving to the right */
	for (wc = w2+1; (wc->word == w2->word) && (wc < w1); wc++) ;
        if (wc >= w1) {
	    wc = w2 ;
	    wordo.offset = wc->offset ;
	}
    }

    /* Allocate a new data block */
    pe = P->adata + P->used ; P->used += 1 ;
    wordo.word = wc->word ;
    *pe = nb = new_data(block->allocated, &wordo, &(block->up)) ;
    nb->apart = P ;
    block->up = wordo ;

    /* Reshuffle the records between the two blocks */
    memcpy(nb->record, wc, (char *)w1 - (char *)wc) ;
    block->used = (wc - w0) ;
    nb->used = (w1 - wc) ;

    /* Reorder the blocks in the partition */
    for (--pe; pe > pd; pe--) pe[1] = pe[0] ;
    pd[1] = nb ;
}

static WORDO *add_rec(word, offset)
/*++++++++++++++++
.PURPOSE  Add a new record in a data block.
.RETURNS  ---
-----------------*/
  char *word ;
  long offset ;
{
  static char *anull = "" ;
  char buf[MAXWORDLEN+1] ;
  char *aword ;
  SDATA **d0, **d1, **d, *sd ;
  WORDO w[2], *wd0, *wd1, *wd ;
  SPART *sp ;
  int   m;

    /* Find first to which block the data has to be inserted 	*/
    if (!spart)  {	/* Initialisation not yet done...	*/
    	w[0].word = anull ; w[0].offset = 0 ;
    	for (m=0; m<sizeof(buf); m++) buf[m] = '\177';
    	buf[sizeof(buf)-1] = 0 ;
    	w[1].word = add_word(buf); w[1].offset = MAX_LOC ;
        spart = new_part(spart_size, w, w+1) ;
        d = spart->adata ;
	w[1].offset = 0 ;
        if (spart_size >= 37) for (m = '0'; m <= 'z'; m++) {
            if (m == '9'+1) m = 'a' ;
            buf[0] = m ; buf[1] = 0 ;
            w[1].word = add_word(buf) ;
            *(d++) = sd = new_data(sdata_size, w, w+1) ;
	    sd->apart = spart ;
	    w[0] = w[1] ;
        }
        *(d++) = sd = new_data(sdata_size, w, &(spart->up)) ;
	sd->apart = spart ;
	spart->used = d - (spart->adata) ;
    }

    sp = spart ;
    while(sp->signature == sign_part) {
	d0 = sp->adata ;
	d1 = d0 + sp->used-1 ;
	while(d0 < d1) {
	    d = d0 + ((d1-d0)/2) ;
	    m = strcmp(word, (*d)->low.word) ;
	    if (m == 0) m = offset - (*d)->low.offset ;
	    if (m < 0) d1 = d;		/* Lower upper limit */
	    else if (d0 == d) {
	        m = strcmp(word, (*d1)->low.word) ;
	        if (m == 0) m = offset - (*d1)->low.offset ;
		if (m >= 0) d0 = d1 ;
		else d1 = d0 ;
	    }
	    else d0 = d ;
	}
	sp = (SPART *)(*d0) ;
    }
    sd = (SDATA *)sp ;
#ifdef DEBUG
    if (strcmp(sd->low.word, sd->up.word) >= 0) {
	fprintf(stderr, "****Problem bad limits(%s/%d, %s/%d)\n",
	sd->low.word, sd->low.offset, sd->up.word, sd->up.offset) ;
    }
    if ((sd->used > 0) && (strcmp(sd->record[0].word, sd->low.word) < 0)){
        fprintf(stderr,
        "****Problem LOW=%s/%d has %s in %d\n", sd->low.word, sd->low.offset,
        sd->record[0].word, 0) ;
    }
    if ((sd->used>0) && (strcmp(sd->record[sd->used-1].word,sd->up.word)>=0)){
        fprintf(stderr,
        "****Problem UP=%s/%d has %s in %d\n", sd->up.word, sd->up.offset,
        sd->record[sd->used-1].word, sd->used-1) ;
    }
#endif

    /* Expand the data block if necessary */
    if (sd->used == sd->allocated) {
	expand_data(sd) ;
	return(add_rec(word, offset)) ;
    }

    /* Find record such that   *rec < wd1	*/
    wd0 = sd->record ;
    wd1 = wd0 + (sd->used-1) ;
    if (wd0 >  wd1) m = -1 ;	/* Empty array	*/
    else m = strcmp(word, wd0->word) ;
    if (m >= 0)  {
	m = strcmp(word, wd1->word) ;
	if (m == 0) m = offset - wd1->offset ;
	if (m >= 0) wd0 = wd1 ;	/* At end	*/
        else while (wd0 < wd1) {
	    wd = wd0 + ((wd1-wd0)/2) ;
	    m = strcmp(word, wd->word) ;
	    if (m == 0) m = offset - wd->offset ;
	    if (m < 0) wd1 = wd ;
	    else if (wd0 == wd) break ;
	    else wd0 = wd ;
	}
    }

    /* Add the word ? */
    if (wd0->word) m = strcmp(word, wd0->word);
    if (m) aword = add_word(word) ; 	/* New Word */
    else {
	aword = wd0->word ;
        m = offset - wd0->offset ;
    }
    if (m == 0) return(wd0) ;		/* Don't add a 2nd time */

    /* Move the records coming after the new one, to make room 	*/
    if (m >= 0) wd0++;		/* Insert AFTER position wd0	*/
    for (wd = sd->record + sd->used; --wd >= wd0; ) wd[1] = wd[0] ;
    sd->used += 1 ;
    /* Place the new entity */
    wd0->word   = aword ;
    wd0->offset = offset ;

    return(wd0) ;
}

static void write_index(sp, outf)
/*++++++++++++++++
.PURPOSE  List all records from partition
.RETURNS  ---
-----------------*/
  SPART *sp ;
  FILE *outf ;
{
  static char *skip = "" ;
  static int recu = 0 ;		/* Recursivity Counter	*/
  static long offset ;
  static char *w ;
  static int pc, np, nd, nw, nq, nf, fmin ;
  SDATA *sd ;
  WORDO *wd1, *wd ;
  int   i ;
  long o ;
    if (!sp) sp = spart ;
    if (recu == 0) {
	np = nd = nw = nq = 0 ;
	/* Flag frequent words	*/
	if (ov) pc = ov, fmin = (ov * articles) / 100 ;
	else pc = 99, fmin = articles ;
	if (fmin < 1) fmin = 1 ;
	w = (char *)0, offset = -300000 ;
	nf = 0 ;		/* Frequency of word    */
	if (ov) printf("....write_index: %6d", 0), fflush(stdout) ;
    }
    recu++ ;
    np++ ;	/* Number of partitions */
    if (ov) printf("\b\b\b\b\b\b%6d", nw), fflush(stdout) ;
    for (i=0; i<sp->used; i++) {
	sd = sp->adata[i];
	if (sd->signature == sign_part) {
	    write_index((SPART *)sd, outf) ;
	    continue ;
	}
	nd++;
	for (wd=sd->record, wd1=wd+sd->used; wd<wd1; wd++) {
	    if (wd->word == skip) continue ;
	    ++nq ;
	    if (wd->word == w) {
	        fprintf(outf, "\t%s", ltoa6(wd->offset)) ;
	        wd->word = skip ;
	        nf++ ;
	        continue ;
	    }
	    if (w) fputc('\n', outf) ;	/* Terminate preceding line */
	    o = ftell(outf) ;
	    nw++ ;
	    if (nf >= fmin) {
        	printf("\n++++Frequency > %d%% '%s' (%d/%d)", pc, w,
		    nf, articles) ;
		if (ov)  printf("\n....write_index: %6d", nw) ;
	    }
	    nf = 1 ;
	    w = wd->word;
	    fprintf(outf, "%s\t%s", w, ltoa6(wd->offset)) ;
	    if ((o-offset) >= BUFSIZE) 	/* Upper Index */
	         wd->offset = offset = o ;
	    else wd->word = skip ;
	}
    }
    --recu ;
    if (recu == 0) {
        fputc('\n', outf) ;
        if (nf >= fmin) {
            printf("\n++++Frequency > %d%% '%s' (%d/%d)", pc, w,
	        nf, articles) ;
            if (ov)  printf("\n....write_index: %6d", nw) ;
        }
        if (ov) printf(
	    "\b\b\b\b\b\b%6dwords, %d quotations, %d blocks, %d partitions.\n",
	    nw, nq, nd, np) ;
    }
}

/*==================================================================
		General Utility Routines
 *==================================================================*/
static char *achar(car)
/*++++++++++++++++
.PURPOSE  Translate a control character
.RETURNS  Equivalence
.REMARKS  \octal or \f \b \r \n ...
-----------------*/
  int car;	  /* IN: Character to translate	*/
{
  static char result[8];

    switch(car) {
    	case '\b': return("\\b");
    	case '\f': return("\\f");
    	case '\n': return("\\n");
    	case '\r': return("\\r");
    	case '\t': return("\\t");
    	case 033: return("\\E");
    }
    if (isgraph(car)) result[0] = car, result[1] = 0;
    else sprintf(result, "\\%03o", car);
    return(result);
}

static int trc(in, out)
/*++++++++++++++++
.PURPOSE  Interpret a string according to C standards
.RETURNS  Number of bytes in out
.REMARKS  \octal or \f \b \r \n ...
-----------------*/
  char *in;       /* IN: String to interpret */
  char *out;      /* OUT: Result of interpretation */
{
  char *p, *q; int o;

   for (p = in, q = out; *p; p++, q++) {
	if (*p == '\\') switch(*++p) {
	    case 'b': *q = '\b'; continue;
	    case 'f': *q = '\f'; continue;
	    case 'n': *q = '\n'; continue;
	    case 'r': *q = '\r'; continue;
	    case 't': *q = '\t'; continue;
	    case 'E':
	    case 'e': *q = 033; continue;
	    default:
		if (isdigit(*p)) {	/* Octal representation */
		    for (o=0; isdigit(*p); p++) o = o * 8 + (*(p++) - '0');
		    *q = o;
		    p--;
		    continue;
		}
	}
	*q = *p;
    }
    *q = 0;
    return (q - out);
}

/*========================================================================
 *              Transform into TeX (option -tex)
 *========================================================================*/

static void pr_tex(text, len, fh)
/*++++++++++++++++
.PURPOSE  Transform the text into TeX
.REMARKS  Presently only ~ (\about) =~ (\simeq)
-----------------*/
  char *text;   /* IN: Text to edit */
  int  len ;	/* IN: Length of .. */
  FILE *fh;     /* IN: File handler */
{
  char *p, *e; int b ;
  static char prec = '\n' ;	/* Preceding char */
    b = strlen(btex) ;
    for (p=text, e=p+len; p<e; prec = *(p++)) switch(*p) {
      case '\\':        /* Already in TeX */
	if (b) { fprintf(fh, "\\%s", btex); btex[b=0] = 0 ; }
        fputc(*p, fh); p++ ;
        if (p < e) fputc(*p, fh) ;
        continue ;
      case '<':
	if (b) {
	    if (b<12) { btex[b++] = *p; btex[b]=0; continue; }
	    fprintf(fh, "\\%s", btex); btex[b=0] = 0 ;
	}
        if (p[1] == '~') fprintf(fh, "{\\la}"), p++;
        else if (p[1] == '=') fprintf(fh, "{\\le}"), p++;
        else fputc(*p, fh) ;
        continue ;
      case '>':
	if (b) {
	    if (b<12) { btex[b++] = *p; btex[b]=0; continue; }
	    fprintf(fh, "\\%s", btex); btex[b=0] = 0 ;
	}
        if (p[1] == '~') fprintf(fh, "{\\ga}"), p++;
        else if (p[1] == '=') fprintf(fh, "{\\ge}"), p++;
        else fputc(*p, fh) ;
        continue ;
      case '%':
	if (b && (b < (sizeof(btex)-2))) {
	    btex[b++] = '\\'; btex[b++] = *p; btex[b] = 0 ;
	    continue ;
	}
	if (prec != '\n') fputc('\\', fh) ;
	fputc(*p, fh) ;
	continue ;
      case '=':
	if (b) {
	    if (b<12) { btex[b++] = *p; btex[b]=0; continue; }
	    fprintf(fh, "\\%s", btex); btex[b=0] = 0 ;
	}
        if (p[1] == '~') fprintf(fh, "{\\simeq}"), p++ ;
        else fputc(*p, fh) ;
        continue ;
      case '~':
        fprintf(fh, "{\\about}");
        continue ;
      case '^': case '_':
	if (b) {	/* End of index / exponent */
	    if (btex[0] == *p) {
		fprintf(fh, "%c{%s}", btex[0], btex+1) ;
		btex[b=0] = 0 ;
		continue ;
	    }
	    /* We have e.g. ^ terminated by _ */
	    fprintf(fh, "\\%s", btex);
	    b = 0 ;
	}
	btex[b++] = *p ;
	btex[b] = 0 ;
	continue ;
      /* DEFAULT: */
      default:
	if (b) {
	    if (isspace(*p)) goto case_put_btex ;
	    goto case_btex ;
	}
        fputc(*p, fh) ;
        continue ;
      case_put_btex:
	fprintf(fh, "\\%s", btex) ;
	btex[b=0] = 0 ;
	fputc(*p, fh) ;
	continue ;
      case_btex:	/* Text saved in btex */
	if (b >= sizeof(btex)-1) goto case_put_btex ;
	btex[b++] = *p ; btex[b] = 0 ;
	continue ;
    }
}


/*==================================================================
		Regular Expressions
 *==================================================================*/

#define INIT		register char *sp = instring ;
#define GETC()		*(sp++)
#define PEEKC()		(*sp)
#define UNGETC(c)	(--sp)
#define RETURN(c)	return(c);
static int compile_errors;
#define	ERROR(c)	++compile_errors, fprintf(stderr, \
    "****Error #%d in regexp '%s'+%d\n", c,instring,(int)(sp-instring))
static int IGNORECASE = 0;
static int GREPword   = 0;
#include <regexp0.h>
#include <regexp1.h>

int stumatch(cexp, input, bos)
/*++++++++++++++++
.PURPOSE  Check whether input satisfies the compiled regular expression
.RETURNS  Number of matched chars / 0 for mismatch / -1 for match negative
-----------------*/
  LINK1 *cexp;	/* IN: Compiled regular expression */
  char *input;	/* IN: String to check	*/
  int bos ;	/* IN: 1 if beginning of string must match */
{
  LINK1 *preg ; int saved_circ, n;
    saved_circ = circf ;
    for (preg = cexp, n=1; preg && n ; preg = preg->previous) {
	circf = bos | preg->circf ;
	if (preg->neg&0100) {		/* FGREP ! */
	    loc1 = preg->neg&040 ?
		 stustr(input, preg->text):	/* ignore case */
		 strstr (input, preg->text);	/* case sensit */
	    if (loc1) n = strlen(preg->text);
	    else n = 0;
	}
	else if (step(input, preg->buf)) n = (loc2-loc1);
        else n = 0;
        if (preg->neg&1) {
	    if (n) n = 0 ;
	    else n = -1 ;
        }
    }
    circf = saved_circ ;
    return(n);
}

int strmatch(rexp, input)
/*++++++++++++++++
.PURPOSE  Check whether input satisfies the regular expression
.RETURNS  Number of matched chars / 0 for mismatch
.REMARKS  Use stumatch for compiled expressions.
-----------------*/
  char *rexp;	/* IN: Regular expression */
  char *input;	/* IN: String to check	  */
{
  static struct {	/* Single-linked list 		*/
	LINK1 *previous;
	char *text;
	char neg ;
	char circf ;
	char buf[BUFSIZ];
  } alink1 ;
    alink1.text = rexp ;
    compile(rexp, alink1.buf, alink1.buf+sizeof(alink1.buf), 0);
    return(stumatch((LINK1 *)&alink1, input, 1));
}

static LINK1 *compile_expr(input, opt)
/*++++++++++++++++
.PURPOSE  Compile a regular expression (starts with /)
.RETURNS  The (newly allocated) compiled expression / NULL
-----------------*/
  char *input;	/* IN: The regular expression, starts by / */
  int opt;	/* IN: EGREP FGREP GREPdelimited */
{
  char buf[BUFSIZ] ;
  char *p, neg;
  int n, delim;
  LINK1 *list;
    /* memset(buf, 0, sizeof(buf)); */
    circf = 0;
    neg = delim = 0 ;
    compile_errors = 0;
    p = input;
    if (opt&GREPdelimited) {
        delim = *p ;
        if ((delim == '!') || (delim == '^') || (delim == '-')) {
	    p++, neg=1 ;
	}
	delim = *(p++);
    }
    if (opt == FGREP) {
	neg = 0100;			/* Indicates fgrep */
	if (IGNORECASE) neg |= 040;
	p = (char *)0;
	n = 0;
    }
    else {
	if (opt&EGREP)
            p = ecompile(p, buf, buf+sizeof(buf), delim);
        else p = compile(p, buf, buf+sizeof(buf), delim);
        if (!p) return((LINK1 *)0) ;
	if (compile_errors) return((LINK1 *)0);
        n = p - buf ;
    }
    list = (LINK1 *)malloc(sizeof(LINK1) + n+2);
    list->text = input ;
    list->previous = (LINK1 *)0;
    list->neg = neg ;
    list->circf = circf ;
    list->word  = GREPword;		/* Implicit \< ... \>	*/
    list->sortsep = 0;
    if (n) {
	p = list->buf;
	if (GREPword) *(p++) = CLT;	/* indicator of \<	*/
	memcpy(p, buf, n);
	if (GREPword) {
	    p[n] = p[n-1];		/* CCEOF (end-of-expr)	*/
	    p[n-1] = CGT;		/* indicator of \>	*/
	}
    }
    return(list) ;
}

static int grepPar(filename, rexp, opt)
/*++++++++++++++++
.PURPOSE  Print to stdout the paragraphs matching a regular expression
.RETURNS  Number of paragraphs matched / -1 = error
-----------------*/
  char *filename; 	/* IN: Input file to read */
  LINK1 *rexp ;		/* IN: Compiled Regular Expression to Check */
  int opt;		/* IN: 1=print text, 2=print lineno, 4=print filename */
{
  char *par, *p;
  FILE *file ; 	/* IN: Input file to read */
  int n = 0 ;
  int lineno = 0 ;

    if ((filename[0] == '-') && (filename[1] == 0))
	file = stdin;
    else file = fopen(filename, "r") ;
    if (!file) {
	fprintf(stderr, "****%s: ", progname) ;
	perror(filename) ;
	return(-1);
    }
    while ((par = getPar(file))) {
	lineno += 1;	/* End-of-Line */
	if (stumatch(rexp, par, 0)) {
	    n++;
	    if (opt&4) { printf("%s", filename); if(opt&3) putchar(':'); }
	    if (opt&2)   printf("%d:", lineno);
	    if (delim[0] && (opt&6)) printf("%s", delim);
	    if (opt&1) printf("%s%s", par, delim);
	}
	for (p=par; *p; p++) { if (*p == '\n') lineno++; }
    }
    if (filename[0] != 0) fclose(file);
    return(n) ;
}

/*==================================================================
		Dealing with Index
 *==================================================================*/
static LINDEX *isearch(lineno, a0, a1)
/*++++++++++++++++
.PURPOSE  Search in Line Index for a given line number
.RETURNS  The LINDEX x such that   x[0].lineno <= lineno < x[1].lineno
.REMARKS  Dichotomic search
-----------------*/
  int lineno;	/* IN: The line number to search */
  LINDEX *a0 ;	/* IN: Starting array of LINDEX  */
  LINDEX *a1 ;	/* IN: --Ending array of LINDEX  */
{
  LINDEX *p0, *p1, *p ;

    if (lineno < a0->lineno) return((LINDEX *)0) ;
    for (p0=a0, p1=a1; p0<p1; ) {
    	p = p0 + (p1-p0)/2;
	if (p == p0) break ;
    	if (lineno < p->lineno) p1 = p ;
	else p0 = p ;
    }
    /* Return a1 in case of mismatch (lineno too high) */
    return(p0) ;
}

static char *gsearch(key, buf, len, recsep, cmp)
/*++++++++++++++++
.PURPOSE  Search in buf line starting with key
.RETURNS  The starting line where   line <= key <= (next line)
.REMARKS  buf[-1] assumed to be a newline
-----------------*/
  char *key;	/* IN: String to find in buf */
  char *buf;	/* IN: Buffer with lines     */
  int  len;	/* IN: Length of buf	     */
  int  recsep;	/* IN: Indicator of EndOfLine*/
  int (*cmp)();	/* IN: Comparison function   */
{
  char *p0, *p1, *p;
#ifdef DEBUG
  int i;

    printf("\n....Look for ") ; putchar('"') ;
    for (p = key; isaword(*p); p++) putchar(*p) ;
    putchar('"') ; putchar('\n') ; x = getchar();
#endif
    if (len <= 0) len = strlen(buf) ;
    for (p0=buf, p1=p0+len; p0<p1; ) {
    	if ((*cmp)(key, p0) <= 0) break;
    	p = p0 + (p1-p0)/2;
#ifdef DEBUG
    	for (i=0; isgraph(p0[i]); i++) ;
    	x=p0[i], p0[i]=0; printf("....p0=%p\"%s\"", p0, p0); p0[i] = x;
    	printf(", p=%p", p);
#endif
    	while (*p != recsep) p-- ; ++p;
#ifdef DEBUG
    	for (i=0; isgraph(p[i]); i++) ;
    	x=p[i], p[i]=0;   printf("=>%p\"%s\"", p, p);  p[i] = x;
#endif
	if (p == p0) {
	    while (*p != recsep) p++;
	    ++p;
	}
#ifdef DEBUG
    	for (i=0; isgraph(p[i]); i++) ;
    	x=p[i], p[i]=0;   printf(" =>%p\"%s\"", p, p);  p[i] = x;
    	printf("; (*cmp)(key,p)=%d.", (*cmp)(key,p));
#endif
    	if ((*cmp)(key, p) >= 0) p0 = p;
    	else {
	    if (p1 == p) {
		p -= 2;
		while (*p != recsep) p--;
		++p;
	    }
	    p1 = p;
	}
    }
    return(p0);
}

static int getNumbers(buf, val, size)
/*++++++++++++++++
.PURPOSE  Interpret a set of POSITIVE Numbers
.RETURNS  Number of values / 0 if Not Numbers
.REMARKS  Unknown numbers are set to 0
		If buf does not start with a digit, line copied to stdout,
		and 0 is returned.
	==> When echo set: display what follows the numbers.
-----------------*/
  char *buf;	/* IN: ascii representation  */
  INT4 *val;	/* OUT: Values	*/
  int  size;	/* IN: Size of val array  */
{
  char *p; int i;
  long atol();

    i = 0;  p = buf;
    while (isspace(*p)) p++;
    if (!isdigit(*p)) {
	printf("%s\n", buf) ;
	/* write(1, buf, strlen(buf));
	  write (1, "\n", 1);
	*/
	return(0);
    }
    while (i < size) {
	while (isspace(*p)) p++;
	val[i++] = atol(p);
	while (isdigit(*p)) p++;
	while (isspace(*p)) p++;
	if (*p == ':') p++;		/* Separator in grep */
    }
    if (ov) printf("....Read %d numbers (%d,%d) from Buffer <%s>\n",
	i, val[0], val[1], buf);
    if (*p && echo) 	/* Edit the input if any (except if :) */
	printf("%s\n", p) ;
    return(i);
}

static int gdif(s1, s2)
/*++++++++++++++++
.PURPOSE  Compare two strings (only ISAWORD, limited to MAXWORDLEN)
.RETURNS  The difference
-----------------*/
  char *s1;	/* IN: String to find in buf */
  char *s2;	/* IN: Buffer with lines     */
{
#if 1
  char *e; int d;
    for (e = s1+maxwordlen; s1<e; s1++,s2++) {
    	if ((!isaword(*s1)) && (!isaword(*s2))) return(0);
	if ((d = (*s1 - *s2))) return (d);
    }
    return(0);
#else
    while(isaword(*s1) && isaword(*s2) && (*s1 == *s2)) s1++, s2++;
    printf(" *s1=%d,*s2=%d", *s1, *s2);
    return (isaword(*s1) || isaword(*s2) ? *s1 - *s2 : 0);
#endif
}

static int ugdif(s1, s2)
/*++++++++++++++++
.PURPOSE  Compare two strings (only ISAWORD), case independent
.RETURNS  The difference
-----------------*/
  char *s1;	/* IN: String to find in buf */
  char *s2;	/* IN: Buffer with lines     */
{
#if 1
  char *e; int d;
    for (e = s1+maxwordlen; s1<e; s1++,s2++) {
    	if ((!isaword(*s1)) && (!isaword(*s2))) return(0);
	if ((d = (tolower(*s1) - tolower(*s2)))) return (d);
    }
    return(0);
#else
    while(isaword(*s1) && isaword(*s2) && (tolower(*s1) == tolower(*s2)))
	s1++, s2++;
    printf(" *s1=%d,*s2=%d", *s1, *s2);
    return (isaword(*s1) || isaword(*s2) ? tolower(*s1) - tolower(*s2) : 0);
#endif
}

static int agdif(s1, s2)
/*++++++++++++++++
.PURPOSE  Compare two strings (only ISAWORD), case independent
.RETURNS  The difference
-----------------*/
  char **s1;	/* IN: String to find in buf */
  char **s2;	/* IN: Buffer with lines     */
{
    return(gdif(*s1, *s2)) ;
}

static int augdif(s1, s2)
/*++++++++++++++++
.PURPOSE  Compare two strings (only ISAWORD), case independent
.RETURNS  The difference
-----------------*/
  char **s1;	/* IN: String to find in buf */
  char **s2;	/* IN: Buffer with lines     */
{
    return(ugdif(*s1, *s2)) ;
}

static int odif(s1, s2)
/*++++++++++++++++
.PURPOSE  Compare two strings (only numeric)
.RETURNS  The difference
-----------------*/
  char *s1;	/* IN: String to find in buf */
  char *s2;	/* IN: Buffer with lines     */
{
    return(a6tol(s1) - a6tol(s2));
}

static int locdif(i1, i2)
/*++++++++++++++++
.PURPOSE  Compare two locations (count mask Position_in_file)
.RETURNS  The difference (opposite)
-----------------*/
  INT4 *i1;	/* IN: First number */
  INT4 *i2;	/* IN: Secnd number */
{
  int i;
    i = (*i2>>MAXWORDS) - (*i1>>MAXWORDS);
    if (i == 0) i = i1[1] - i2[1];
    return(i);
}

static int pardif(par1, par2)
/*++++++++++++++++
.PURPOSE  Compare two paragraphs
.RETURNS  The difference (opposite)
.REMARKS  Sort order in sort_keys
-----------------*/
  char **par1;	/* IN: First paragraph */
  char **par2;	/* IN: Secnd paragraph */
{
  LINK1 *ps; int dif = 0;
  char *p1, *p2;
  int iw, sep;
    for (ps = sort_keys; (dif == 0) && ps; ps = ps->previous) {
	p1 = parloc(*par1, ps->text); if (!p1) return(256);
	p2 = parloc(*par2, ps->text); if (!p2) return(256);
	iw = ps->word;
	sep = ps->sortsep;
	if ((iw==0) && ps->circf)	/* Numeric sort -- need next field */
	    iw = 1;
	while (--iw>=0) {
	    if (sep) {
		while (*p1 && (*p1 != sep) && (*p1 != '\n')) p1++;
		while (*p2 && (*p2 != sep) && (*p2 != '\n')) p2++;
		if (*p1 == sep) p1++;
		if (*p2 == sep) p2++;
	    }
	    else {
	        while (isgraph(*p1)) p1++;
	        while ((*p1 == ' ') || (*p1 == '\t')) p1++;
	        while (isgraph(*p2)) p2++;
	        while ((*p2 == ' ') || (*p2 == '\t')) p2++;
	    }
	}
	if (ps->circf)       dif = atoi(p1) - atoi(p2);	/* Numeric (-n) */
	else if (IGNORECASE) dif = stucmp(p1, p2);	/* foldcase(-f) */
	else                 dif = strcmp(p1, p2);
	if (ps->neg) dif = -dif;
    }
#ifdef DEBUG
    printf("%4d ", dif);
    for (iw=9; --iw>=0 && isprint(*p1); p1++) putchar(*p1);
    printf(" / ");
    for (iw=9; --iw>=0 && isprint(*p2); p2++) putchar(*p2);
    putchar('\n');
#endif
    return(dif);
}

static void fload(fname)
/*++++++++++++++++
.PURPOSE  Load the "Word Index"  (-f option)
.RETURNS  ---
.REMARKS  buf[-1] assumed to be a newline
-----------------*/
  char *fname;	/* IN: File name (-f arg.)   */
{
  int len, i, j, nw;
  char *p, *pC, *pK, *n;
  char **aK, **a ;

    if (dic_file) {
	if (strcmp(dic_file, fname) == 0) return;	/* Already loaded */
	free(dic_file), dic_file = (char *)0;
    }
    if (dic_cont) free(dic_cont), dic_cont = (char *)0;

    /* Open the new file */
    i = open(fname, 0|O_BINARY);
    if (i <= 0) { perror(fname); exit(1); }
    len = lseek(i, 0L, 2);	/* File Size */
    lseek(i, 0L, 0);
    dic_file = strdup(fname);
    dic_cont = malloc(len+1);
    if((j=read(i, dic_cont, len))!=len) fprintf(stderr, 
	    "#+++got %d/%d bytes for: %s\n", j, len, fname); 
    dic_cont[len] = 0;
    close(i);
    if (ov) printf("....Read file of %d bytes: %s\n", len, fname);

				/* Load #Keep  and   #Change */
    pC = pK = (char *)0 ;
    for (p = dic_cont; *p; p = n ) {
	if (*p == '\n') p++;
	n = p + strloc(p, '\n');
	if (*p != '#') continue;
	*(p++) = 0;
	if (strncmp(p, "Keep:", 5) == 0)
	    pK = n;
	else if (strncmp(p, "Change:", 6) == 0)
	    pC = n;
	else continue;
    }
    if (pC) {
	word_change = pC;
	i = strlen(pC) - 1;
	while (isspace(pC[i])) i--;
	lword_change = ++i;
    }
    if (pK) {
	word_keep = pK;
	i = strlen(pK) - 1;
	while (isspace(pK[i])) i--;
	lword_keep = ++i;
    }
    if (pC) {
        nw = 0 ;	/* Count number of words (at right of =) to keep */
	for (p = word_change, n = word_change + lword_change; p < n; p++) {
	    if (!isaword(*p)) { 		/* Comment ? */
		p += strloc(p, '\n') ;
		continue ;
	    }
	    i = strloc(p, '\n') ;
	    j = strloc(p, '=') ;
	    if (i < j) { p += i ; continue ; }
	    p += j ; if (p >= n) continue ;
	    while (*p != '\n') {
		if (!isaword(*p)) { p++ ; continue ; }
		nw++;
		while (isaword(*p)) p++ ;
	    }
	}
	/* .. Add the words to keep ... */
	if (pK) for (p = word_keep, n = word_keep + lword_keep; p < n; p++) {
	    if (!isaword(*p)) { 		/* Comment ? */
		p += strloc(p, '\n') ;
		continue ;
	    }
	    nw++ ;
	    while (isaword(*p)) p++ ;
	}
	aK = a = (char **)malloc((nw+1)*(sizeof(char *))) ;

	/* Set up the pointers to the words, to sort them */
	for (p = word_change, n = word_change + lword_change; p < n; p++) {
	    if (!isaword(*p)) { 		/* Comment ? */
		p += strloc(p, '\n') ;
		continue ;
	    }
	    i = strloc(p, '\n') ;
	    j = strloc(p, '=') ;
	    if (i < j) { p += i ; continue ; }
	    p += j ; if (p >= n) continue ;
	    while (*p != '\n') {
		if (!isaword(*p)) { p++ ; continue ; }
		*(a++) = p ;
		while (isaword(*p)) p++ ;
	    }
	}
	if (pK) for (p = word_keep, n = word_keep + lword_keep; p < n; p++) {
	    if (!isaword(*p)) { 		/* Comment ? */
		p += strloc(p, '\n') ;
		continue ;
	    }
	    *(a++) = p ;
	    while (isaword(*p)) p++ ;
	}
	*a = (char *)0 ;

	/* Sort the equivalences */
	qsort (aK, nw, sizeof(char *), (SORT_FCT)agdif) ;

	/* Count the number of different things to keep */
	for (a=aK, n = "", len=0; *a; a++) {
	    if (gdif(n, *a)) {
		n = *a, len++ ;
		for (p=n; isaword(*p); p++) len++ ;
	    }
	}

	/* Copy the new Keep table */
	pK = word_keep = malloc(len+2) ;
	*(pK++) = '\n' ;
	for (a=aK, n = ""; *a; a++) {
	    if (gdif(n, *a)) {
		n = *a;
		for (p=n; isaword(*p); p++) *(pK++) = *p ;
		*(pK++) = '\n' ;
	    }
	}
	*pK = 0 ;
	lword_keep = pK - word_keep ;
	free(aK) ;
    }
}


/*==================================================================
		Open/Close Parfiles
 *==================================================================*/
static int CloseParfile(pf)
/*++++++++++++++++
.PURPOSE  Close the Parfile
.RETURNS  0 / -1
-----------------*/
  PARFILE *pf;	/* MOD: File Number */
{
	/* Open the database */
    if (!pf->db) return(0);	/* Already Closed */
    if (pf->gf) {		/* Associated General Index */
	if (close(pf->gf)) perror(pf->name);
	if (pf->gindex) free(pf->gindex-1);
	if (pf->agindex) free(pf->agindex-1);
    }
    if (close(pf->db)) {	 /* Error closing  */
  	perror(pf->name);
	exit(1);
    }
    opened -= 1;
    pf->db = pf->gf = pf->rank = 0;
    pf->lgindex = pf->lagindex = 0;
    if (pf->delim) free(pf->delim);
    if (pf->idb)  free(pf->idb);
    pf->idx = pf->ide = (LINDEX *)0;
    return(0);
}

static int OpenParfile(pf)
/*++++++++++++++++
.PURPOSE  Open the Parfile
.RETURNS  0 / -1
-----------------*/
  PARFILE *pf;	/* MOD: File Number */
{

		/* Open the database */
    if (pf->db) { /* Already opened !  */
	pf->rank = ++orank;
	return(0);
    }

		/* Close files if required */
    if (opened >= MAXOPENED) {
    	PARFILE *p; int i;
	p = (PARFILE *)0;
	for (i=0; i < parfiles; i++) {
	    if (parfile[i].rank == 0) continue;
	    if (!p) p = &parfile[i];
	    if (p->rank > parfile[i].rank) p = &parfile[i];
	}
	CloseParfile(p);
    }

		/* Open the new file */
    /*printf("#...open(%s,%d)\n", pf->name, O_RDONLY|O_BINARY); */
    pf->db = open(pf->name, 0|O_BINARY);
    if (pf->db <= 0) {
	perror(pf->name);
	exit(1);
    }
    pf->rank = ++orank;
    pf->idb = pf->delim = (char *)0;
    pf->idx = pf->ide = (LINDEX *)0;
    if (ov) printf("....OpenParfile %s (%d)\n", pf->name, pf->db);

    /* V5.22: add the title of the parfile */
    if (addIntro&2) { struct stat fst; int i; 
	char host[32]; /* V5.51: add the hostname */
	gethostname(host, sizeof(host));
	fstat(pf->db, &fst);
	printf("%%#Parfile(%s)@%s on: %s(%s)\n",
		VERSION, host, pf->name, adate(fst.st_mtime));
	printf("%%#Execution date: ");
	i = strlen(pf->name) + strlen(VERSION) -1;
	while (--i>=0) putchar(' ');
	printf("%s\n", adate(time(0)));
	addIntro = 1;
    }
    if (addIntro&1) {
	putchar('\n');
	addIntro = 0;
    }
    opened += 1;
    return(0);
}

/*==================================================================
		Dealing with Line Index
 *==================================================================*/
static int OpenLindex(pf, mode)
/*++++++++++++++++
.PURPOSE  Open the Lindex (.i0)
.RETURNS  File Number / -1
-----------------*/
  PARFILE *pf;	/* MOD: File Number */
  int  mode;	/* IN: file Mode     */
{
  int dbi;

    OpenParfile(pf);
    strcpy(buf, pf->name); strcat(buf, ".i0");
    if (mode) mode |= O_TRUNC|O_CREAT;
    dbi = open(buf, mode|O_BINARY, 0644);
    if ((dbi <= 0) && (mode == 1))	/* Write */
	dbi = creat(buf, 0644);
    if (dbi <= 0) perror(buf);

    return(dbi);
}

static int LoadLindex(pf)
/*++++++++++++++++
.PURPOSE  Load into memory the Lindex (.i0)
.RETURNS  0 / -1
-----------------*/
  PARFILE *pf;	/* MOD: File Number */
{
  int dbi, stat, len, align;
  char *p;
  LINDEX *pi;

    dbi = OpenLindex(pf, 0);
    if (dbi <= 0)	return(dbi);

    len = lseek(dbi, 0L, 2);	/* File Size */
    lseek(dbi, 0L, 0);

    /* We have to align the Line Index on a correct boundary */
    for (align=0; len%sizeof(INT4); align++, len++) ;
    len -= align;
    pf->idb = malloc(len+align);
    p = pf->idb + align;
    pf->idx = (LINDEX *)p; pf->ide = (LINDEX *)(p+len);

    while (len > 0) {
  	stat = read(dbi, p, len);
	if (stat < 0)	perror(pf->name);
	if (stat == 0)
	    fprintf(stderr, "*** parfile: bad index of %s\n", pf->name);
	if (stat <= 0)	return(-1);
	len -= stat; p += stat;
    }
    close(dbi);

    /* Set pf->idx (skip first line) */
    p = pf->idb + align;
    if (strncmp(p, "Parfile.", 8) == 0) {
	stat = strloc(p, '\n'); p[stat] = 0;
	pf->idx = (LINDEX *)(p+stat+1);
	if (align) strcpy(pf->idb, p);
    	/* Retrieve the Delimiter if known 	*/
	for (p=pf->idb; *p; p++) { char x;
	    if (*p != ' ')  continue ;
	    if (strncmp(p, " -d", 3)) continue;
	    p += 3;
	    len = strloc(p, ' '); x = p[len]; p[len] = 0;
	    if (pf->delim) free(pf->delim);
	    delim = pf->delim = strdup(p); p[len] = x;
	    trc(delim, delim);
	    break;
	}
    }

    /* Reorganize index in  Host local binary */
    if (htonl(123456789L) != 123456789L)
       for (pi = pf->idx; pi < pf->ide; pi++) {
	pi->lineno = ntohl(pi->lineno);
	pi->offset = ntohl(pi->offset);
    }

    return(0);
}

/*==================================================================*/
static int CreateLindex(pf)
/*++++++++++++++++
.PURPOSE  Create an index (.i0)
.RETURNS  Number of lines
.REMARKS  Close everything
-----------------*/
  PARFILE *pf;	/* MOD: File Number */
{
  LINDEX index, netix;
  long  o;	/* Offset of read buf in File */
  int   dbi, len, lexp, onl, i, err;
  char  *p0, *p1, *p2;

    dbi = OpenLindex(pf, 1);	/* Create */
    if (dbi <= 0) 	return(-1);

    /* Write the Header */
    strcpy(buf, "Parfile.i0(4) -d");
    p0 = buf + strlen(buf);
    for (p1=delim; *p1; p1++) strcpy(p0, achar(*p1)), p0 += strlen(p0);
    *(p0++) = '\n';
    if(write(dbi, buf, (p0-buf))<0) { perror(pf->name); exit(1); }

    lexp = strlen(delim);
    index.offset = index.lineno = 0;
    o = 0;
    p0 = p1 = p2 = buf;
    onl = 1;		/* Assume a newline as previous character */

	/* p0 = beginning, p1 = end, p2 = max to search for */
    while(1) {
	if (p2 > p1) p2 = p1;
	len = p1 - p2;	/* Remaining bytes in buffer */
	bcopy(p2, buf, len); p0 = buf + len;
	o += (p2-buf);
	len = read(pf->db, p0, sizeof(buf)-len);
	if (len < 0) { perror(pf->name); exit(1); }
	if (len == 0)	break;
	p1 = p0+len; p2 = p1 - lexp;
	p0 = buf;
	while (p0 < p2) {
	    if (onl) {
		index.lineno += 1;
		onl = 0;
	    	if (strncmp(p0, delim, lexp) == 0) {
		    for (i=lexp; --i >= 0; p0++) {
			if (*p0 == '\n') index.lineno += 1;
		    }
		    index.offset = o + (p0-buf);
		    netix.lineno = htonl(index.lineno),
		    netix.offset = htonl(index.offset);
		    if (lexp != 0) err = write (dbi, &netix, sizeof(netix))
			!= sizeof(netix) ;
		    else err = write(dbi, &(netix.offset), sizeof(INT4))
			!= sizeof(INT4);
		    if (err) {
			perror("****Writing Lindex");
			exit(1);
		    }
		}
		continue;
	    }
	    onl = *(p0++) == '\n';
	}
    }
    index.lineno += 1;
    index.offset = o + (p0-buf);
    netix.lineno = htonl(index.lineno),
    netix.offset = htonl(index.offset);
    if (lexp != 0) len = write (dbi, &netix, sizeof(netix)) ;
    else len = write(dbi, &(netix.offset), sizeof(INT4)) ;
    if (len<0) { perror(pf->name); exit(1); }
    close(dbi);
    if (ov) printf("....Created Index file: %s.i0\n", pf->name);
    CloseParfile(pf);

    return(0);
}

/*==================================================================
		Dealing with General Index
 *==================================================================*/

static int LoadGindex(pf)
/*++++++++++++++++
.PURPOSE  Load into memory the General Index (.i-)
.RETURNS  0 / -1
-----------------*/
  PARFILE *pf;	/* MOD: File Number */
{
  int len;
  INT4 o;
  char line[LENLINE1+2], *p, x;

    if (pf->gf) return(0);
    OpenParfile(pf);
    strcpy(buf, pf->name); strcat(buf, xtension);
    pf->gf = open(buf, O_BINARY);
    if (pf->gf < 0) {
	pf->gf = 0;
	perror(buf);
	return(-1);
    }
    /* Read first line, which is
    	"\tParfile.i-(4) main_index_size ... -ddelim (options)"
    */
    len = read(pf->gf, line, sizeof(line)-1); line[len] = 0;
    len = strloc(line, '\n'); line[len] = 0;
    for (p=line; isspace(*p); p++) ;
    oc1 = inumop = optii = opti2 = 0;
    while (isgraph(*p)) {		/* Check also .i-%	*/
	if (*p == '%') oc1 = '%';
    	p++;
    }
    while (isspace(*p)) p++;
    o = a6tol(p);			/* Size of main Index 	*/
    while(isgraph(*p)) p++;

    /* Retrieve the Options, separated by Tabs	*/
    p += strloc(p, '\t');
    while(*p) {
	if (*p != '\t') { p += strloc(p, '\t'); continue ; }
	if (strncmp(p, "\t-f ", 4) == 0) { 	/* Dictionary	*/
	    p += 4;
	    len = strloc(p, '\t');
	    x = p[len]; p[len] = 0;
	    fload(p);
	    p[len] = x;
	    p += len;
	    continue;
	}
	if (strncmp(p, "\t-lw ", 5) == 0) { 	/* Word Length	*/
	    p += 5;
	    maxwordlen = atoi(p) ;
	    if (maxwordlen < 2) maxwordlen = MAXWORDLEN ;
	}
	if (strncmp(p, "\t-d", 3) == 0) { 	/* Delimiter	*/
	    p += 3;
	    for (len = 0; isgraph(p[len]); len++) ;
	    /* len = strloc(p, '\t'); */
	    x = p[len]; p[len] = 0;
	    if (pf->delim) free(pf->delim);
	    delim = pf->delim = strdup(p); p[len] = x;
	    trc(delim, delim);
	    p += len;
	    continue;
	}
	if (strncmp(p, "\t-ii", 4) == 0) { 	/* Keep lower	*/
	    p += 4;
	    optii = 1 ;
	    continue ;
	}
	if (strncmp(p, "\t-i1", 4) == 0) { 	/* Keep short	*/
	    p += 4;
	    opti2 = 1 ;
	    continue ;
	}
	if (strncmp(p, "\t-i2", 4) == 0) { 	/* Keep short	*/
	    p += 4;
	    opti2 = 2 ;
	    continue ;
	}
	if (strncmp(p, "\t-inum", 6) == 0) { 	/* Delimiter	*/
	    p += 6 ;
	    inumop = 1 ;
	    continue ;
	}
	p++;
    }

    len = lseek(pf->gf, 0L, 2) - o;	/* Size of SuperIndex */
    p = malloc(len+2);
    *(p++) = '\n';			/* Start with newline */
    pf->gindex = p;
    lseek(pf->gf, (long)o, 0);
    pf->lgindex = read(pf->gf, pf->gindex, len);
    /* Remove the emtpy lines, if any	*/
    while (((*(pf->gindex) == '\r') || (*(pf->gindex) == '\n'))
         && (pf->lgindex > 0)) pf->gindex+= 1, pf->lgindex -= 1 ;
    if (pf->lgindex <= 0) {
	perror("Bad or Emtpy General Index");
	return(-1);
    }
    pf->gindex[pf->lgindex] = 0 ;
    if (ov) fprintf(stderr, "----Loaded Gindex of %d bytes:\n%s\n",
        pf->lgindex, pf->gindex) ;
    return(0);
}

static int LoadIndex2(pf, key)
/*++++++++++++++++
.PURPOSE  Load into memory the Index part containing key
.RETURNS  0 / -1
-----------------*/
  PARFILE *pf;	/* MOD: File Number */
  char *key;	/* IN: Superindex part  "key\tOffset\nkey\toffset */
{
  int len;
  INT4 o;
  char *p;

    if (pf->agindex) {		/* Already loaded. Is it the good one ? */
	if (gdif(key, pf->agindex) == 0) return (0); 	/* Yes, it is ! */
    }

	/* Compute the size which is required */
    for (p=key; isgraph(*p); p++) ; p++;	/* Skip TAB */
    o = a6tol(p); while(isgraph(*p)) p++; p++;	/* Skip \n  */
    while(isgraph(*p)) p++; p++;		/* Skip TAB */
    len = a6tol(p) - o;

	/* Allocate a buffer if necessary */
    if (len >= pf->lagindex) {
	if (pf->agindex) free(pf->agindex-1);
	pf->lagindex = 1 + ((1+len)|0777);	/* Round    */
	pf->agindex = malloc(pf->lagindex);
	*(pf->agindex) = '\n'; pf->agindex += 1;
    }

	/* Read the Index File */
    if (lseek(pf->gf, (long)o, 0) == -1) {
	perror("****Seeking the .i- index file");
	return(-1);
    }
    len = read(pf->gf, pf->agindex, len);
    if (len < 0) {
	perror("****Reading the .i- index file");
	return(-1);
    }
    pf->bagindex = len;		/* Actual Length */

    return(0);
}

static char *theword(str, len, letter)
/*++++++++++++++++
.PURPOSE  Standardize the word to keep in General Index
.RETURNS  NULL (word to ignore) / The word / (starts by '=' for a substitution)
.REMARKS  All in lowercase
-----------------*/
  char *str;	/* IN: The word to standardize */
  int  len;	/* IN: its length */
  int  letter;	/* IN: %-flag: 1 means "Keep" */
{
  static LINK1 *americanize ;
  static char local[40] ;
  static char word[MAXWORDLEN+1];
  char *p, *e, *eword, *s, x;
  int  k, umlaut, mac, mod;

#ifdef DEBUG
    printf("....theword(") ;
    fwrite(str, 1, len, stdout) ;
    printf(", letter='%c')\n", letter) ;
    if (ov) {
        printf("....theword(len=%d \"", len) ;
	fflush(stdout) ;
        fwrite(str, 1, len, stdout) ;
        printf("\", letter=%02x)\n", letter) ;
    }
#endif

    eword = word + maxwordlen ;		/* Max size of a word */
    /* Should the word be kept? according to leading letter.. */
    if (ogk && letter) {
	k = strloc(ogk, letter);
	if (*ogk == '-') {
	    if (ogk[k]) return((char *)0);
	}
	else if (!ogk[k]) return((char *)0);
    }

    mod = 0 ;				/* Any modification -> No replacement */

    /* Ignore leading non-alphanumerical */
    if ((len>0) && (!isalnum(*str))) str++, len-- ;

    e = str + len;
    /* Ignore the word if only numeric	 */
    if ((inumop == 0) && isdigit(*str))  {
	for (p=str; (p<e) && isdigit(*p); p++) ;
	if (p >= e) return((char *)0);
    }
    if (len == 0) return((char *)0);	/* Empty Word */

    /* Before making all modifications, copy word to a local buffer 	*/
    if (!americanize) americanize
        = compile_expr("/.*[iy]s[ei][sd]*$/", GREPdelimited) ;
    if (len > sizeof(local)-2) len = sizeof(local)-2 ;
    strncpy(local, str, len) ; str = local ;
    str[len] = 0 ;
    if (oc&4) for (p=local; *p; p++) *p = tolower(*p) ;

    /* We change the ending 'ies' into 'y'	*/
    if ((len>4) && (len<(maxwordlen+3)) && (strcmp(str+len-3, "ies")==0)) {
	len -= 2 ;
	str[len] = 0 ;
	str[len-1] = 'y' ;
	mod++ ;
    }

    /* We remove the ending 's' if:
	   - not an ending 'ss', 'as', 'is', 'os' or 'us'
    */
    e = str + len - 1;
    if ((len>3) && (/*tolower*/(*e) == 's')) switch(/*tolower*/(e[-1])) {
	default:
	    x=e[1], e[1]=0;
	    p = gsearch(str, word_keep+1, lword_keep-1, '\n', ugdif);
	    e[1]=x;
	    if (ugdif(p, str)) mod++, str[--len] = 0;
	    break;
	case '\'':	/* Remove 's */
	    len -= 2;
	    mod++ ;
	    break;
	case 's': case 'a': case 'i': case 'o': case 'u':
	    break;
    }

    /* Set up the "americanisation rule" : ending ise --> ize 	*/
    /* e.g. analyze analyzis	*/
    if ((len > 4) && stumatch(americanize, str, 1)) {
	e = str + len - 2 ;
	if (*e != 's') e-- ;
	if (*e == 's') mod++, *e = 'z' ;
    }

    /* k is set to 1 if word includes capitals other than the first letter */
    /* if (len > MAXWORDLEN) len = MAXWORDLEN; */
    mac = (len > 6) && (strncasecmp(str, "mac", 3) == 0) ;
    k = isupper(*str) ? -1 : 0 ;
    /* Umlaut: not for last letter (i.e. scan only to s+len-1),
       and require a word of at least 4 letters */
    umlaut = 0;
    /* Copy also to word -- and change if umlaut rule */
    if (len>3) {
	for (p=word, s=str, e=s+len-1; (s<=e) && (p<eword); s++) {
	    if (isupper(*s)) k += 1;
    	    *p = tolower(*s);
	    if ((p != word) && (p[-1] == p[0])) {	/* Dble letter 	*/
	        if (*p == 'a') { mod++; continue ; }	/* aa --> a	*/
	        if (*p == 'u') { mod++; continue ; }	/* uu --> u	*/
	    }
	    if ((*p == 'e') && (p != word) && (p!=e) && isalpha(s[1])) {
	        /* Umlaut, remove 'e'	*/
	        if ((p[-1] != 'o') && (p[-1] != 'a') && (p[-1] != 'u')) p++;
	        else mod++, umlaut++;
	    }
	    else if (mac && (p == &word[1]))
		mod++, mac = 0;				/* Mac --> Mc	*/
	    else p++ ;
        }
        *p = 0;
    }
    else {
	for (p=word, s=str, e=s+len; (s<e) && (p<eword); s++)
	     *(p++) = tolower(*s);
	*p = 0;
    }
    if (umlaut) {	/* Verify Word to keep */
	x = *s; *s = 0;
	p = gsearch(str, word_keep+1, lword_keep-1, '\n', ugdif);
	*s = x;
	if (ugdif(p, str)) mod++;	/* Removing the 'e' was correct */
	else {
	    for (p=word, s=str, e=s+len; (s<e) && (p<eword); s++)
	    	*(p++) = tolower(*s);
	    *p = 0;
	    mod -= umlaut ;
	}
    }

    if (letter=='A') {
	return(len >= 2 ? word : (char *)0);	/* Author: Keep */
    }

    /* Keep words which have more than 1 capital letter */
    if ((len <= 2) && isalpha(*word) && (k == 0)) {
        if (opti2 && opti2 <= len) ;
	else {
	    p = gsearch(word, word_keep+1, lword_keep-2, '\n', gdif);
	    if (gdif(word, p)) return ((char *)0); /* NOT in word_keep */
	}
    }

    if (mod) return(word) ;	/* A modifcation --> Don't skip !!! */

    p = gsearch(word, word_change+1, lword_change-3, '\n', gdif);
    if (gdif(word, p) == 0) {
	while(isaword(*p)) p++;
	if (*p == '=') {	/* Take a synonym */
	    /* if (ov) printf("....Word %s replaced by ", p); */
	    return(p);
	    /* for (++p, e=word; isaword(*p); p++, e++) *e = *p;
	    *e = 0;
	    if (ov) printf("%s\n", word);
	    return(word); */
	}
	return((char *)0);
    }
    return(word);
}

static int match_ascword(str, len)
/*++++++++++++++++
.PURPOSE  Get a full word, ascii option
.RETURNS  Matched bytes / 0 (can't decide) / -1 (not a valid word)
.REMARKS  The latex symbols are ignored; result saved in word_abuf
-----------------*/
  char *str ;	/* Input text */
  int len ;	/* str length */
{
  char *p, *e, *w ;
    p = str, e = str + len ;
    if (p >= e) return(0) ;
    if (!isaword(*p)) return(-1) ;
    for (w = word_abuf; isaword(*p) && (p<e) && (w<word_ebuf) ; p++)
        *(w++) = *p ;
    *w = 0 ;
    if (p >= e) return(0) ;	/* Not enough bytes, can't decide...	*/
    return(p-str) ;
}

static int match_texword(str, len)
/*++++++++++++++++
.PURPOSE  Get a full word, latex option
.RETURNS  Matched bytes / 0 (can't decide) / -1 (not a valid word)
.REMARKS  The latex symbols are ignored; result saved in word_abuf
-----------------*/
  char *str ;	/* Input text */
  int len ;	/* str length */
{
  char *p, *e, *m, *w, *eq ;
  int brace, c ;
    p = str, e = str + len ;
    if (p >= e) return(0) ;
    if (!isawtex(*p)) return(-1) ;
    for (w=word_abuf, brace=0; (p<e) && (w<word_ebuf) ; p++) {
        switch(*p) {
	  case '{': brace++; continue ;
	  case '}': brace--; continue ;
	  case '\\':
	    p++ ;
	    if (p >= e) return(0) ;
	    if (!isalpha(*p)) continue ;	/* Ignore accent */
	    m = p ;
	    while (isalpha(*p) && (p < e)) p++ ;
	    c = *p; *p = 0 ;
	    eq = gsearch(m, tex_symbols, -1, '\n', gdif) ;
	    if (gdif(eq, m)) eq = 0 ;
	    *p = c ;
	    if (eq) {
		eq += strloc(eq, '=') ;
		for (++eq; *eq != '\n'; eq++, w++) *w = *eq ;
	    }
	    else	/* Default Value of Macro = its value... */
		/* while (isalpha(*m)) *(w++) = *(m++) */ ; /*NONE*/
	    while ((*p == ' ') && (p < e)) p++ ;
	    --p ;
	    continue ;
	}
	if (isalnum(*p)) *(w++) = *p ;
	else break ;
    }
    *w = 0 ;
    if (p >= e) return(0) ;	/* Not enough bytes, can't decide...	*/
    if (brace) return(-1) ;
    return(p-str) ;
}

static int CreateGindex(pf)
/*++++++++++++++++
.PURPOSE  Create General Index of an Opened File
.RETURNS  0 / -1
.REMARKS  The header is  Parfile.i0-%..(4).
		If delimiter is empty, only the pffset of each line is saved.
-----------------*/
  PARFILE *pf;		/* IN: Parfile */
{
  /* static char ext[] = ".i-"; */
  static char word[120];
  char *pw, *sw;
  FILE *outf ;
  char *p0, *p1, *p2, *p, c;
  int lexp, len, eof, i, ignore_line, n;
  int nq = 0 ;
  char cpc, onl, opc;
  INT4 o , offset ;
  LINK1 *preg;

    /* Initialize the word_abuf for word_fct */
    word_abuf = word ; word_ebuf = word_abuf + sizeof(word) - 1 ;

    OpenParfile(pf);
    if (ov) {
	printf("....Words discarded:") ;
	for (i = '\n', p = word_change; *p; p++) {
	    if (i == '\n') putchar('\t') ;
	    i = *p ;
	    if (i != '\n') putchar(i) ;
	}
	putchar('\n') ;
    }

    /* Verify Recursive Definitions in Dictionary... */
    n = 0 ;
    for (p0=word_change+1, p1=p0+lword_change-1; p0<p1; p0++) {
	p0 += strloc(p0, '=') ;
	if (p0 >= p1) continue ;
	while (*p0 != '\n') {
	    while (isascii(*p0) && (!isaword(*p0))) p0++ ;
	    if (p0 >= p1) break ;
	    p2 = gsearch(p0, word_change+1, lword_change-1, '\n', ugdif) ;
	    if (p2 && (ugdif(p0, p2) == 0)) {	/* Redefinition ! */
		/* A redefinition may occur due to cutting words,
		   e.g. catalogue=catalog in case of 7 letters maxi.
		*/
		if (p0 == (p2 + strloc(p2, '=') + 1)) ;
		else {
		    if ((n == 0) && dic_file) fprintf(stderr,
		        "++++In dictionary: %s\n", dic_file) ;
		    n++ ;
		    for (pw = p0; isgraph(*pw); pw++) ;
		    eof = *pw ; *pw = 0 ;
		    fprintf(stderr, "++++Word '%s' recursively translated: ",
			p0) ;
		    *pw = eof ;
		    for (pw = p0; *pw != '\n'; pw--) ; ++pw ;
		    fwrite(pw, 1, strloc(pw, '\n'), stderr) ;
		    fprintf(stderr, "; ") ;
		    fwrite(p2, 1, strloc(p2, '\n'), stderr) ;
		    fprintf(stderr, "\n") ;
		}
	    }
	    while (isaword(*p0) && (p0<p1)) p0++ ;
	}
    }
    if (n) fprintf(stderr, "++++%d recursive translation(s) found\n", n) ;

    onl = 1;	/* Assume newline as previous char */
    opc = 0;	/* Set to 1 if % just read */
    cpc = 0;
    pw = word;
    o = offset = articles = 0;
    p0 = p1 = p2 = buf;
    lexp = strlen(delim);

    if (ov) printf("....create_index: %9d", 0), fflush(stdout) ;
    eof = 0; ignore_line = 0;
    while(!eof) {
	if (p2 > p1) p2 = p1;
	len = p1 - p2;	/* Remaining bytes in buffer */
	bcopy(p2, buf, len); p0 = buf + len;
	o += (p2-buf);
	len = sizeof(buf) - len - 1;	/* Max to read */
	len = read(pf->db, p0, len) ;
#ifdef DEBUG
    	printf("\n....Input file: read %d bytes (lexp=%d)\n", len, lexp);
#endif
	if (len < 0) { perror(pf->name); exit(1); }
	if (len == 0)	eof = 1;
	p1 = p0+len; *p1 = 0;
	p2 = eof ? p1 : p1 - lexp;
	p0 = buf;
	while (p0 < p2) {
#ifdef DEBUG
	    printf("...buf+%d: \"", (p0-buf)) ;
	    for (i=0; i<20 && (p0+i)<p1; i++) {
		if (isprint(p0[i])) putchar(p0[i]) ;
		else printf("\\%03o", p0[i]) ;
	    }
	    putchar('"') ;
	    if ((p0+i)<p1) printf("...") ;
	    printf("\tonl=%d, ig_line=%d, opc=%d, cpc=%d, oc1=%c (HR)",
		onl, ignore_line, opc, cpc, oc1) ;
	    getchar() ;
#endif
	    if (onl) {		/* Previous char was a newline */
	    	if (strncmp(p0, delim, lexp) == 0) {
		    if (lexp) onl = p0[lexp-1] == '\n';
		    else onl = 0;
		    p0 += lexp;
		    offset = o + (p0-buf);
		    articles ++ ;	/* New Article	*/
		    continue ;
		}
		if (((p2-p0) < lreg) /* Not enough bytes read 	*/
		    && (!eof)) { p2 = p0; break; }
		onl = 0;
		if ((opc = *p0 == oc1)) { p0++; continue; }

		/* Look: line in list of lines to ignore / keep ? */
		ignore_line = 0;
		if ((preg = reg_leave)) {
		    ignore_line = stumatch(preg, p0, 1) ;
		    /* while (preg && (!ignore_line))
		    	ignore_line = stumatch(preg, p0, 1),
			preg = preg->previous;
		    */
		}
		if ((!ignore_line) && (preg = reg_keep)) {
		    ignore_line = 1;
		    ignore_line = (stumatch(preg, p0, 1) == 0) ;
		    /* while (preg && ignore_line)
		    	ignore_line = (stumatch(preg, p0, 1) == 0),
			preg = preg->previous;
		    */
		}
		continue;
	    }

	    if (opc) {		/* Previous char is a '%' */
		opc=0, cpc = *p0;
		if (isgraph(*p0)) { p0++; continue; }
	    }

	    if ((ignore_line) || (articles == 0)) {  /* V5.49: ignore par#0 */
		onl = *(p0++) == '\n'; 
		continue ; 
	    }

	    n = (*word_fct)(p0, p2-p0) ;
	    if (n == 0)  { p2 = p0; break ; }	/* Too short ... */
	    if (n < 0) {
	        onl = *(p0++) == '\n';
		continue ;
	    }

	    /* I've got n bytes from input stream, word saved in word_abuf */
	    p0 += n ;
	    sw = theword(word, strlen(word), cpc);
	    if (!sw) continue ;		/* Word to Ignore */
	    if (*sw == '=') sw++;
	    /* Equivalence: several words when a blank */
	    while(isgraph(*sw)) {
	        for (p=sw; isgraph(*p); p++) ;
	        c = *p ; *p = 0 ;
	        add_rec(sw, offset) ;
	        ++nq ;
	        if (ov && ((nq%5000) == 0)) printf("\b\b\b\b\b\b\b\b%8d", nq),
	    	    fflush(stdout) ;
	        *p = c ;
	        sw = p ;
	        if (c == ' ') sw++ ;
	    }
	}
    }
    if (ov) printf(
	"\b\b\b\b\b\b\b\b\b%9d (non-squeezed) quotations for %d articles\n",
	nq, articles);
#ifdef DEBUG
    printf("....    Index created in memory\n") ;
#endif

		/* Create the Index File */
    strcpy(buf, pf->name); strcat(buf, xtension);
    outf = fopen(buf, "w+");
    if (! outf) { perror(buf); exit(1); }
    /* Write the Header ...  */
    for (o=0; o<LENLINE1; o++) buf[o] = ' ';
    buf[LENLINE1] = 0;
    fprintf(outf, "%s\n", buf);
    fflush(outf) ;
#ifdef __MSDOS__
    buf[120] = 0;
#endif

    /* Write the Main Index, and write in Index the offset of Index of Index */
    write_index((SPART *)0, outf) ;
    fflush(outf);
    o = ftell(outf); 		/* End of Main Index */

    /* Rewrite the Parfile Header */
    fseek(outf, 0L, 0);
    pw = ogk ? ogk-1 : "";
    sprintf(buf, "\tParfile%s%s(4) %s bytes(64-base) of main index:\t-d",
	xtension, pw, ltoa6(o));
    p0 = buf + strlen(buf);
    for (p1=delim; *p1; p1++) strcpy(p0, achar(*p1)), p0 += strlen(p0);
    if (opti2){ strcpy(p0, "\t-i"); p0 += strlen(p0) ; *(p0++) = '0'+opti2; }
    if (optii) strcpy(p0, "\t-ii"), p0 += strlen(p0) ;
    if (inumop) strcpy(p0, "\t-inum"), p0 += strlen(p0) ;
    if (itexop) strcpy(p0, "\t-itex"), p0 += strlen(p0) ;
    if (maxwordlen != 7) sprintf(p0, "\t-lw %d", maxwordlen), p0 += strlen(p0) ;
    if (dic_file) {
	p1 = "" ;
	if (*dic_file != '/') {
	    p1 = malloc(LENLINE1-3) ;
	    if (!getcwd(p1, LENLINE1-4)) perror("****"), strcpy(p1, ".") ;
	    strcat(p1, "/") ;
	}
	sprintf(p0, "\t-f %s%s", p1, dic_file), p0 += strlen(p0);
	if (*p1) free(p1) ;
    }
    if ((preg = reg_leave)) {
	while (preg)
	    sprintf(p0, "\t-i!%s", preg->text),
	    p0 += strlen(p0),
	    preg = preg->previous ;
    }
    if ((preg = reg_keep)) {
	while (preg)
	    sprintf(p0, "\t-i%s", preg->text),
	    p0 += strlen(p0),
	    preg = preg->previous ;
    }
    *(p0++) = '\t';
    *p0 = 0;
    buf[LENLINE1] = 0;
    fprintf(outf, "%s", buf);
    fflush(outf);

    /* Append SuperIndex */
    fseek(outf, o, 0);
    write_index((SPART *)0, outf) ;
    fflush(outf);
    fprintf(outf, "~~~~~~~\t%s\n", ltoa6(o));
    fclose(outf);

    return(0);
}

/*==================================================================
		Display Parfile's Extracts
 *==================================================================*/
static int DisplayBytes(pf, offset, expr)
/*++++++++++++++++
.PURPOSE  Display until specified expression
.RETURNS  Number of bytes displayed
.REMARKS  NO DISPLAY if the same paragraph as before
-----------------*/
  PARFILE *pf;	/* IN: File concerned */
  INT4 offset;	/* IN: Offset in file */
  char *expr;	/* IN: Expression of article separator */
{
  INT4 len;
  int  lexp, bytes, onl;
  char *p0, *p1, *p2, goon ;

    OpenParfile(pf);	/* Will increase the "rank" */
    if (ov) printf(".... DisplayBytes of File %s, offset %d, expr %s\n",
  	pf->name, offset, expr);
    if (lseek(pf->db, (long)offset, 0) < 0) { perror(pf->name); return(0); }

    lexp = strlen(delim);
    p0 = p1 = p2 = buf;
    onl = 0;		/* Assume a newline NOT as previous character */
    goon = 1;		/* Continue indicator */
    btex[0] = 0 ;	/* No saved tex text  */
    if (o_n) {
	printf("%s", o_n) ;
	if (multiple) printf("%d:", thefileno) ;
	printf("%d\n", offset) ;
	/*
	write(1, o_n, strlen(o_n));
	sprintf(charo, "%ld\n", offset);
	write(1, charo, strlen(charo));
	*/
    }

	/* p0 = beginning, p1 = end, p2 = max to search for */
    bytes = 0 ;
    while(goon) {
	if (p2 > p1) p2 = p1;
	if (otex) pr_tex(buf, p2-buf, stdout) ;
	else fwrite(buf, p2-buf, 1, stdout) ;
	bytes += p2-buf ;
	/* write(1, buf, p2-buf); */
	len = p1 - p2;	/* Remaining bytes in buffer */
	bcopy(p2, buf, len); p0 = buf + len;
	len = read(pf->db, p0, sizeof(buf)-len);
	if (len < 0) { perror(pf->name); exit(1); }
	if (len == 0)	break;
	p1 = p0+len; p2 = p1 - lexp;
	p0 = buf;
	while (p0 < p2) {
	    if (onl) {
		onl = 0;
	    	goon = strncmp(p0, delim, lexp);
	    	if (!goon) break;
		p0 += lexp;
	    }
	    else onl = *(p0++) == '\n';
	}
    }
    if (otex) pr_tex(buf, p0-buf, stdout);
    else fwrite(buf, p0-buf, 1, stdout);
    bytes += p0-buf ;
    /* write(1, buf, p0-buf); */
    return(bytes);
}

static int Display(afileno, lineno)
/*++++++++++++++++
.PURPOSE  Display the paragraph
.RETURNS  Number of bytes displayed
.REMARKS  NO DISPLAY if the same paragraph as before
-----------------*/
  INT4 afileno;	/* IN: File number to display */
  INT4 lineno;	/* IN: Line number to display */
{
  PARFILE *pf;
  static LINDEX *last = (LINDEX *)(-1);
  static LINDEX *x    = (LINDEX *)0;
  static int   last_fileno = -1;
  int len, i, ii, bytes, current_line;
  INT4 *offset;
  char *p=0, *abuf, *ebuf, errmsg[80];

    thefileno = afileno ;	/* For DisplayBytes */
    if ((afileno < 0) || (afileno >= parfiles)) {
	fprintf(stderr, "*** parfile: Invalid File Number %d not in [0,%d]\n",
		afileno, parfiles-1);
	return(0);
    }
    pf = &parfile[afileno];
    if (obyte)	return(DisplayBytes(pf, lineno, delim));
    if (!pf->idx) {
  	if (LoadLindex(pf) < 0)	exit(-1);
  	x = (LINDEX *)0;
  	last = (LINDEX *)(-1);
    }
    if (last_fileno != afileno) {
  	last_fileno = afileno;
  	x = (LINDEX *)0;
  	last = (LINDEX *)(-1);
    }
    pf->rank = ++orank;
    if (!x)	      x = pf->idx;
    if (x>= pf->idx)  x = pf->idx;

    /* A single array of offsets of line beginnings is immediate ! */
    if (!*delim) {
	offset = (INT4 *)pf->idx + (lineno-1);
	if (offset <  (INT4 *)pf->idx) return(0);
	if (offset >= (INT4 *)pf->ide) return(0);
	return(DisplayBytes(pf, *offset, pf->delim));
    }
    /* while ((x > pf->idx) && (x->lineno >  lineno)) x--; */
    /* while ((x < pf->ide) && (x->lineno <= lineno)) x++; */
    if (lineno < x->lineno) {
	x = isearch(lineno, pf->idx, x) ;
    }
    else {
	x = isearch(lineno, x, pf->ide) ;
    }

    if (x >= pf->ide)	return(0);
    if (x == last)	return(0);	/* Same as previous */

    if (squeeze)	last = x; /* Test compared to previous is removed... */

    if (o_n) {
	strcpy(errmsg, o_n); p = errmsg + strlen(errmsg);
	if (afileno > 1) sprintf(p, "%d:", afileno), p += strlen(p);
    }
    if (!x) 	{	/* First paragraph (updated V5.46) */
	len = pf->idx ? pf->idx[0].offset : 0;
	lseek(pf->db, 0L, 0); current_line = 0 ;
	if (o_n) strcpy(p, "0\n");
    }
    else {
	len  = x[1].offset - x[0].offset ;
	lseek(pf->db, (long)x->offset, 0);
	current_line = x->lineno ;
	if (o_n) sprintf(p, "%ld\n", (long)x->offset);
    }
    if (o_n) printf("%s", errmsg); /* write(1, errmsg, strlen(errmsg)); */

    bytes = 0;
    btex[0] = 0 ;
    while (len > 0) {
	i = MIN(len, sizeof(buf));
	if ((len-i) == 1)	i--;
	ii = read(pf->db, buf, i);
	if (ii < (i-1)) {
	    sprintf(errmsg, "*** parfile: read  %d bytes out of %d", ii, i);
	    perror(errmsg);
	}
	len -= i; abuf = buf ; ebuf = abuf + i ;
	/* Remove End of Paragraph */
	if ((len == 0) && (i > 1) && (ebuf[-2] == '\n') && (ebuf[-1] == '\n'))
	    ebuf-- ;
	if (inMerge) while ((abuf<ebuf) && (current_line < lineno)) {
	    if (*abuf == '\n') current_line += 1;
	    abuf++ ;
	}
	i = ebuf - abuf ;
	if (otex) pr_tex(abuf,i,stdout) ;
	else fwrite(abuf,i,1,stdout);
	/* write(1,abuf,i); */
	bytes += i;
    }
    if (btex[0]) printf("%s", btex) ;
    return(bytes);
}

static char* toupperString(char *in) {
    char *p,*out = strdup(in);
    for (p=out; *p; p++) *p=toupper(*p);
    return out;
}

static int DisplayKey(pkeys)
/*++++++++++++++++
.PURPOSE  Display the paragraphs matching the set of keys
.RETURNS  Number of bytes displayed
.REMARKS  The input buffer (keys) is modified
-----------------*/
  char *pkeys;	/* IN: Set of keys separated by spaces */
{
  PARFILE *pf;
  char *pw[MAXWORDS], *paw[MAXWORDS], *p, *pi, x;
  char *keys, *kupper=0, *words=0, *w;
  INT4 minw, mask, locs[MAXWORDS], *alocs, *plocs;
  INT4 ow_print = 0;		/* Edition of weight */
  int len, nf, i, o, nw, difo, imin, found, bytes, lwords=0;
  static char *author = "if not an author's name";

    /* Be sure the Index is loaded... */
    if (LoadGindex(parfile)) exit(1);

    /* Convert keys to lowercase, and truncate */
    while (isspace(*pkeys)) pkeys++;
    keys = strdup(pkeys);
    for (p=keys, len=0; *p; p++) {
	if (optii) *p = tolower(*p);
	if (isaword(*p)) /* *p = tolower(*p) */ ;
	else *p = ' ';
    }

    /* Check the words, issue a message if word normally removed
       lword is non-zero when word(s) are to be replaced by synonyms
    */
    /*lwords = 0; words = (char *)0;*/
    for (p=keys; *p; ) {
	while (isspace(*p)) p++;
	o = (p-keys);	/* V5.53: Keep position in list of words */
	len = strloc(p, ' ');
	pi = theword(p, len, 0);
	if (!pi) {
	    x = p[len], p[len] = 0;
	    if ((og&(SEARCH_SILENT|1)/*3*/) == 1) fprintf(stderr,
		"++++The word '%s' is discarded %s\n", p,
		oc1 ? author : "");
	    p[len] = x;
	}
	else {
	    if (ugdif(pi, p) && ((og&(SEARCH_SILENT|1)/*3*/) == 1)) { 
	      int i; char save;
	    	x = p[len], p[len] = 0;
		i = strloc(pi, '\n'); save = pi[i]; pi[i] = 0;
	    	fprintf(stderr,
		  "++++The word '%s' should be replaced by '%s' %s\n",
	          p, pi + (*pi == '='), author);
		pi[i] = save;
	        p[len] = x;
	    }
	    if (og & SEARCH_SILENT) {	/* Verify word in uppercase? */
                lwords += 1 + strloc(pi, (*pi == '=' ? '\n' : 0)); /* V5.54 */
		if (!kupper) kupper = toupperString(pkeys);
                if (gdif(kupper+o,pkeys+o)) 
                    /* do not use the singular */
                    lwords += 1;
            }
	}
	/* while (--len >= 0) *p = tolower(*p), p++ ; */
	p += len ;
	while (isspace(*p)) p++;
    }

    /* Silent replacement: allocate a copy with substituted words */
    if (lwords) {
	w = words = malloc(lwords + strlen(keys));	/* Mod. V5.54 */
	p=keys; while(isspace(*p)) p++;
	while(*p) {
	    len = strloc(p, ' ');
	    if ((pi = theword(p, len, 0))) {
	        if (w != words) *(w++) = ' ';
		if (*pi == '=') pi++;
	        while (*pi && (*pi != '\n')) *(w++) = *(pi++);
	    }
	    p += len;
	    while (isspace(*p)) p++;
	}
	*w = 0;
	w = keys ; keys = words ; words = w;
    }
    /* Set everything to lowercase */
    for (p=keys; *p; p++) *p = tolower(*p) ;

    /* Examine in turn all indexes */
    bytes = 0;
    for(pf=parfile, nf=parfiles; --nf >= 0; pf++) {
	if (LoadGindex(pf)) exit(1);
	/* For each word of the list, keep in the paw array the
	 * matching lines of the index (containing the locations
	 * of the actual paragraphs)
	 */
       	for (p=w=keys, nw=0, found=0; *p; p=w, nw++) {
	    /* ALL words must be found, unless SEARCH_BEST is specified */
	    if ((nw != found) && (!(og&SEARCH_BEST))) break;
	    if (found >= MAXWORDS) {
		fprintf(stderr, "++++ Too many words, ignored =>%s\n", p);
		found = MAXWORDS;
		break;
	    }
	    /* Set up w to the next word */
	    for (w=p; isaword(*w); w++); while (isspace(*w)) w++;

	    pw[found] = paw[found] = (char *)0;
    	    pi = gsearch(p, pf->gindex, pf->lgindex, '\n', gdif) ;
	    if (gdif(p, pi) < 0) continue;	/* Can load the Index */
	    if (LoadIndex2(pf, pi)) continue;	/* Can't load Index2  */
	    pw[found] = gsearch(p, pf->agindex, pf->bagindex, '\n', gdif);
	    if (gdif(p, pw[found])) continue;	/* No Match */
	    /* printf("#...pw[%d]=%p %dbytes\n", 
	       found, pw[found], strlen(pw[found])); */
    	    pw[found] = paw[found] = strdup(pw[found]);
	    /* Each pw gives the position in the file */
	    while (isaword(*pw[found])) pw[found] += 1;
	    while (isspace(*pw[found])) pw[found] += 1;
	    found++;
	  /*_Next:*/
	    /* Move to next word */
    	    while (isaword(*p)) p++;
    	    while (isspace(*p)) p++;
    	}
	if (!found)	/* No word could be found... */
	    continue;

	/* In SEARCH_BEST option, create the association
	   FileLocation ==> Number of Words in the paragraph at this location
	*/
	if (og&SEARCH_BEST) {

	    /* Get the first locations of each word, and
	       set imin to the smallest of these locations
	    */
	    for (i=imin=0; i<found; i++) {
	        locs[i] = a6tol(pw[i]);
	        if (locs[i] < locs[imin]) imin = i;
	        while (isgraph(*pw[i])) pw[i] += 1;
	        if (*pw[i] == '\t') pw[i] += 1;
	    }

	    /* Count how many locations are needed */
	    for (nw=i=0; i<found; i++) {
		nw += 2;	/* 1 TAB + 1 number already read */
	        for (p=pw[i]; *p != EOL; p++) {
	            if (*p == '\t') nw++;
		}
	    }
	    /* Reserve 2 integers for each word occurence, and keep
	       Nocc (5 bits) Mask (MAXWORDS bits) Location in file
	    */
	    plocs = alocs = malloc((nw+1)*2*sizeof(INT4));

	    /* Fill the alocs arrays with the (sorted) contents */
	    while (locs[imin] < MAX_LOC) {
		plocs[0] = 0;
		plocs[1] = locs[imin];
		for (i=imin; i<found; i++) {
		    if (locs[i] != plocs[1]) continue;
		    plocs[0] += (1<<MAXWORDS);
		    plocs[0] |= (1<<(MAXWORDS-1-i));
		    locs[i] = a6tol(pw[i]) ;	/* Get Next Value */
		    while (isgraph(*pw[i])) pw[i] += 1;
		    if (*pw[i] == '\t') pw[i] += 1;
		}
		/* Find the New Minimum */
		for (imin=0, i=1; i<found; i++) {
		    if (locs[i] < locs[imin]) imin = i;
		}
		plocs += 2;
	    }
	    nw = (plocs - alocs)/2;
	    plocs[0] = plocs[1] = 0;

	    /* Sort against the frequency */
	    qsort(alocs, nw, 2*sizeof(INT4), (SORT_FCT)locdif);

	    /* Edit the Results -- stop according to oglist limit */
	    if (oglist <= 0) minw = 0;
	    else minw = ((alocs[0]>>MAXWORDS) - (oglist-1)) << MAXWORDS;
	    for (plocs=alocs; *plocs; plocs += 2) {
		if (plocs[0] <= minw) break;
		if (o_w && (ow_print != plocs[0])) {
		/* Write out the Weight: in silent mode (-gs),
		   issue the weighty only when it's changed.
		*/
		    ow_print = plocs[0];
		    printf("%s %d ", o_w, plocs[0]>>MAXWORDS);
		    x = '('; mask = 1<<(MAXWORDS-1);
		    for (i=0; i<found; ++i, mask>>=1) {
			if (!(plocs[0]&mask)) continue;
			putchar(x);
			for (p=paw[i]; isgraph(*p); p++) putchar(*p);
		        x = ' ';
		    }
		    printf(")\n");
		}
	        bytes += DisplayBytes(pf, plocs[1], delim) + 1;
		printf("%s", delim);
	    }
	    if ((og&SEARCH_SILENT)==0) ow_print = 0;

	    /* Clean up... */
	    free(alocs);
	    goto free_Arrays;
	}

	/* Here (not in SEARCH_BEST option) ALL words must be found. */
    	if (nw != found)
	    goto free_Arrays;

	/* Find the required paragraphs, and display those */
    	while (*pw[0] != EOL) {
    	    for (i=1, difo=0; i<nw && difo==0; i++) {
    		while (*pw[i] != EOL) {
    	    	    difo = odif(pw[0], pw[i]);
    	    	    if (difo <= 0) break;
    	    	    while (isgraph(*pw[i])) pw[i] += 1;
	    	    if (*pw[i] == '\t') pw[i] += 1;
	    	}
    		if (*pw[i] == EOL) difo = -1;
    	    }
    	    if (difo == 0) {
    	    	bytes += DisplayBytes(pf, a6tol(pw[0]), delim) + 1;
		printf("%s", delim) ;
		/* write(1, delim, strlen(delim)); */
	    }
    	    while (isgraph(*pw[0])) pw[0] += 1;
	    if (*pw[0] == '\t') pw[0] += 1;
    	}
      free_Arrays:
	for (i=0; i<found; i++) free(paw[i]);
    }

    if (words)  free(words);
    if (kupper) free(kupper);
    free(keys);
    return(bytes);
}

/*==================================================================*/

static int Merge(name)
/*++++++++++++++++
.PURPOSE  Merge input file with paragraphs extracted from parfile
.RETURNS  0 (OK) / -1
.REMARKS  It's merely a Join...
-----------------*/
  char *name;	/* IN: File name to copy */
{
  char *p, *p1, *ps, inbuf[8192], *line ;
  int  in, len, lineno, mul2;
  INT4 val[3];

    squeeze = 0; mul2 = multiple + 2;
    in = open(name, 0|O_BINARY);
    if (in <= 0) { perror(name); return(-1);}

    inMerge = 1;
    lineno = len = 0; 	p1 = ps = p = inbuf;
    btex[0] = 0 ;

    while((line = Gets())) {
	if (!getNumbers(line, val, mul2)) continue;
				/* [0] = Where, [1] = par-index */
	if (val[0] <= lineno) {
	    fprintf(stderr, "\n*** parfile: Input not sorted =>%s\n",
	    line); inMerge = 0; return (-1);
	}
	if (!multiple) val[2] = val[1], val[1] = 0;
	while (lineno < val[0]) {	/* Copy input to output */
	    if (len <= 0) {
	        if (otex) pr_tex(p1, ps-p1, stdout);
	        else fwrite(p1, ps-p1, 1, stdout);
	        /* write(1, p1, ps-p1); */
	        for (p1 = inbuf; ps < p; p1++, ps++)	*p1 = *ps;
	        p = p1;
	        p1 = ps = inbuf;
	        len = read(in, p, sizeof(inbuf) - (p-inbuf));
	        if (len < 0) { Perror; inMerge = 0; return(-1);}
	        if (len == 0)  { inMerge = 0; return(0); }
	    }
	    while (--len >= 0) {
	        if (*(p++) == '\n') {
	    	    lineno++;
	    	    if ( lineno < val[0]) ps = p;
	    	    break;
	        }
	    }
	}
	if (!orm)  ps = p;
	if (otex) pr_tex(p1, ps-p1, stdout);
	else fwrite (p1, ps-p1, 1, stdout);
	/* write (1, p1, ps-p1); */
	p1 = p;
	Display(val[1], val[2]);	/* Suppress the blank line */
    }

	/* Write out now the remaining of file */
    if (otex) pr_tex(p, len, stdout);
    else fwrite (p, len, 1, stdout);
    /* write (1, p, len); */
    while(1) {
	len = read(in, inbuf, sizeof(inbuf));
	if (len < 0) { Perror; inMerge = 0; return(-1);}
	if (len == 0)  break;
	if(otex) pr_tex(inbuf, len, stdout);
	else fwrite(inbuf, len, 1, stdout);
	/* write (1, inbuf, len); */
    }
    close(in); inMerge = 0;
    return(0);
}

static int merge_files(merge, numerge, ofile)
/*++++++++++++++++
.PURPOSE  Merge files (in parsort)
.RETURNS  0 (OK) / -1
.REMARKS  Files read in parallel.
-----------------*/
  MERGEFILE *merge;	/* IN: Files to merge */
  int numerge;		/* IN: How many files */
  FILE *ofile;		/* IN: Output file    */
{
  int i, imin;

    /* Read first paragraph of all files */
    for (i=0; i<numerge; i++) readPar(merge+i);

    /* Issue first paragraph of 1st file */
    fprintf(ofile, "%s%s", merge[0].par, delim);

    /* Read second paragraph of all files */
    for (i=0; i<numerge; i++) readPar(merge+i);

    while (1) {
	/* Find first element */
	for (imin=0; (imin<numerge) && feof(merge[imin].file); imin++) ;
	if (imin >= numerge) break;	/* All files exhausted */
	for (i=imin+1; i<numerge; i++) {
	    if (pardif(&(merge[i].par), &(merge[imin].par)) < 0)
		imin = i;
	}
	fprintf(ofile, "%s%s", merge[imin].par, delim);
	readPar(merge+imin);		/* Get the next record */
    }
    return(0);
}

/*==================================================================*/

int pargrep(int argc, char **argv, int xgrep)
/*++++++++++++++++
.PURPOSE  Main
.RETURNS  Status (0=OK, 1=nothing found, 2 = errors)
.REMARKS  xgrep=2 for egrep
-----------------*/
{
static char usage[] = "\
Usage: pargrep [-dexpr] [-i] [-E] [-F] [-v] [-p] [-q] [-n] [-l] [-e] regexpr\n\
       parfile... ";
static char help[] = "\
   -dexpr: Define delimiter expression (anchored at line beginning)\n\
           -d alone stops at end-of-line; default is paragraph\n\
           (may contain C-like controls e.g. \\n)\n\
       -i: Make the search case-insensitive\n\
       -E: Use full regular expression\n\
       -F: Use a string comparison\n\
       -v: Give all paragraphs NOT satisfying to the regular expression\n\
       -q: Quiet mode, nothing printed -- just set the return code\n\
       -l: Quiet mode, print just filenames having matches\n\
       -n: print line numbers of found paragraphs\n\
       -w: the regular expression must be in a word (emulates \\<...\\>)\n\
       -e: specifies the regular expression (e.g. when starting by a -)\n\
       -p: no header line in the output (plain).\n\
  parfile: file(s) made of paragraphs\n\
The output consists of paragraphs from parfile, preceded by a 2-lines header;\n\
the header may be removed with the -p option.\n\
" ;
  int opt, neg, status, count, gopt;
  char *p, *pattern, **args;
  LINK1 *rexp;
  char *file;

    /* parprog specifies which type of grep:
       0 = fgrep ; 1 = grep ; 2 = egrep */
    /* Take care of Options */
    pattern = (char *)0;
    status = 1;
    rexp = (LINK1 *)0;
    file = (char *)0;
    neg = 0;
    opt = 1;			/* Default: print the matching */
    gopt = xgrep;
    args = argv;
    while (--argc > 0) {
	p = *++argv;
	if (*p == '-') {
	  CheckOpt:		/* Option 	*/
	    switch(p[1]) {
	    case 'd':			/* Delimiter	*/
		delim = p+2;
		trc(delim, delim);
		continue;
	    case 'e':
		if (p[2]) pattern = p+2;
		else { argc--; pattern = *++argv; }
		continue ;
	    case 'E':
		gopt = EGREP;
		break;
	    case 'F':
		gopt = FGREP;
		break;
	    case 'h':
	    	if (strcmp(p, "-help") == 0) {
		    printf("%s\n%s", usage, help);
		    return(0);
		}
		goto bad_option;
	    case 'i':			/* All indexes	*/
	    	IGNORECASE = 1;
		break;
	    case 'n':			/* Print line number */
		opt |= 2;
		break;
	    case 'l':
		opt = 256;		/* Quiet, just filenames */
		break;
	    case 'p':
		addIntro = atoi(p+2);	/* Plain, no first paragraph */
		while (isdigit(p[2])) p++;
		break;
	    case 'q':
		opt = 0;		/* Quiet, don't print */
		break;
	    case 'v': 			/* Non-matching */
		if (strncmp(p, "-ver", 4) == 0) {
		    printf("%s (CDS) Version %s\n", progname, VERSION);
		    return(0);
		}
		neg = 1;
		break;
	    case 'w': 			/* Words-only	*/
		GREPword = 1;
		break;
	    default:
	    bad_option:
	    	fprintf(stderr, "****pargrep: Invalid Option: %s\n%s\n",
		  p, usage);
	    	return(2);
	    }
	    if (p[2]) { p++; goto CheckOpt; }
	    continue;
	}

	if (!pattern) {
	    pattern = p;
	    continue;
	}

	/* Here I've a file name */
	file = p;
	rexp = compile_expr(pattern, gopt);
	if (neg) rexp->neg ^= 1;
	if (!rexp) return(2);
	/* Output option: print also filename (4) if several files */
	if ((opt&7) && (argc>1)) opt |= 4;
	/* Write the introduction paragraph */
	if (args && *args && (opt&1) && addIntro) {
	    if (addIntro&2) {
	        printf("%%#%c%s", toupper(progname[0]), progname+1);
	        for (++args, count=0; *args; count++, args++) {
		    if (count == 10) printf(" ...");
		    else if (count<10) printf(" %s", *args);
	        }
	        printf("\n%%#Execution date: %s\n", adate(time(0)));
	        addIntro = 1;
	    }
	    if (addIntro&1) { putchar('\n'); addIntro = 0; }
	}
	else printf("\n");		/* Just an empty line... */
	count = grepPar(file, rexp, opt);
	if (count<0) return(2) ;	/* An error occured  */
	if (count>0) {			/* At least 1 result */
	    status = 0;
	    if (opt&256) 		/* Option -l (list files) */
		printf("%s\n", file);
	}
    }

    if (!pattern) {			/* Missing pattern... */
	fprintf(stderr, "****pargrep: Missing pattern\n%s\n", usage);
	return(2);
    }
    if (!file) {			/* Nothing executed, use stdin */
	file = "-";
	rexp = compile_expr(pattern, gopt);
	if (!rexp) return(2);
	if (neg) rexp->neg ^= 1;
	if (args && *args && (opt&1) && addIntro) {
	    if (addIntro&2) {
	        printf("%%#%s", progname);
	        for (++args; *args; args++) printf(" %s", *args);
	        printf("\n%%#Execution date: %s\n", adate(time(0)));
	        addIntro = 1;
	    }
	    if (addIntro&1) { putchar('\n'); addIntro = 0; }
	}
	else printf("\n");		/* Just an empty line... */
	count = grepPar(file, rexp, opt);
	if (count<0) return(2);
	if (count>0) status = 0;
    }
    return(status);
}

int parindex(int argc, char **argv, int arg1)
/*++++++++++++++++
.PURPOSE  Main
.RETURNS  Status
.REMARKS  xgrep=2 for egrep
-----------------*/
{
    fprintf(stderr, "*****Program not yet accessible: %s\n",
      progname);
    return(1);
}

int parsort(int argc, char **argv, int arg1)
/*++++++++++++++++
.PURPOSE  Main
.RETURNS  Status
.REMARKS  xgrep=2 for egrep
-----------------*/
{
static char usage[] = "\
Usage: parsort [-dexpr] [-i] [-f] [-n] [-r] [-u] [-m] [-o output] \n\
       [-k[+word[sep]] line_key] parfile ";
static char help[] = "\
   -dexpr: Define delimiter expression (anchored at line beginning)\n\
           -d alone stops at end-of-line; default is paragraph\n\
           (may contain C-like controls e.g. \\n)\n\
       -f: fold case -- identical to -i\n\
       -i: Make the sort case-insensitive\n\
       -m: merge input files, assumed to be sorted\n\
       -n: sort in numerical order\n\
       -r: reverse the order (i.e. decreasing sort)\n\
       -u: sort unique: remove duplicated paragraphs\n\
       -o: define the output file (default is standard output)\n\
       -k: define the sort keys in 2 parts, line definition and displacement:\n\
           - word indicates the displacement in the line, as a word offset\n\
	     (word#0 is the key which introduces the line)\n\
	     word may be followed by a separator e.g. \n\
	     -k+2= means text between 2nd and 3rd equal sign in line\n\
           - line_key is the beginning of line of sort key\n\
  parfile: input file to sort\n\
The output contains the sorted paragraph (the first one however remains at \n\
the top of the file). If no key is specified, the beginning of each paragraph\n\
is used.\n\
" ;
  static LINK1 ekey= { (LINK1 *)0, };
  static MERGEFILE *merge;
  static int numerge;		/* How many files in merge */
  int num, neg, count, i, uniq;
  char *input, *ofile, *p;
  char **pars, **a, **b;
  LINK1 *rexp, *last;
  FILE *of;

    /* Take care of Options: here 01=numeric, 02=uniq, 0100=rev */
    rexp = last = (LINK1 *)0;
    input = ofile = (char *)0;
    uniq = neg = num = 0;	/* Indicates reverse order     */
    while (--argc > 0) {
	p = *++argv;
	if (*p == '-') {
	  CheckOpt:		/* Option 	*/
	    switch(p[1]) {
	    case 'd':		/* Delimiter	*/
		delim = p+2;
		trc(delim, delim);
		continue;
	    case 'f':
	    case 'i':
		IGNORECASE = 1;
		break;
	    case 'h':
	    	if (strcmp(p, "-help") == 0) {
		    printf("%s\n%s", usage, help);
		    return(0);
		}
		goto bad_option;
	    case 'm':			/* merge option   */
		if (input) {		/* BEFORE files!  */
		    fprintf(stderr, "**** -m BEFORE files!\n%s", usage);
	    	    return(1);
		}
		merge = (MERGEFILE *)malloc(MAXFILES*sizeof(merge));
    	   	memset(merge, 0, MAXFILES*sizeof(merge));
		continue;
	    case 'n':			/* Numerical sort */
		num = 1;
		break;
	    case 'o':			/* Output file    */
		if (p[2]) p += 2;
		else { p = *++argv; --argc; }
		ofile = p;
		continue;
	    case 'r':			/* Reverse order */
		neg = 1;
		break;
	    case 'u': 			/* Unique sort */
		uniq = 1;
		break;
	    case 'v': 			/* Version... */
		if (strncmp(p, "-ver", 4) == 0) {
		    printf("%s (CDS) Version %s\n", progname, VERSION);
		    return(0);
		}
		goto bad_option;
	    case 'k':			/* SORT KEYS */
		--argc;
		rexp = compile_expr(argv[1], FGREP);
		if (!rexp) return(1);
		rexp->circf = num;
		rexp->neg  = neg;
		/* V5.48: accept delimiter */
		p += 2; rexp->word = atoi(p);	/* Word number */
		if (*p == '+') p++;
		while (isdigit(*p)) p++;
		if (*p) rexp->sortsep = *p;
		++argv;
		if (last) last->previous = rexp;
		else sort_keys = rexp;
		last = rexp;
		continue;
	    default:
	    bad_option:
	    	fprintf(stderr, "****%s: Invalid Option: %s\n%s\n",
		  progname, p, usage);
	    	return(2);
	    }
	    if (p[2]) { p++; goto CheckOpt; }
	    continue;
	}

#if 0	/* Accept several files -- these a just concatenated */
	if (input) {
	    fprintf(stderr, "****%s: only 1 input file allowed\n%s\n",
		    progname, usage);
	    return(1);
	}
#endif
	if (merge) {
	    if (numerge >= MAXFILES) {
		fprintf(stderr, "****Too many files in parsort -m\n%s", usage);
		return(1);
	    }
	    merge[numerge].file =
	        (p[0] == '-') && (p[1] == 0) ?	/* file name is - */
		stdin : fopen(p, "r");
	    if (!merge[numerge].file) {
		fprintf(stderr, "****parsort -m: ");
		perror(p);
		return(1);
	    }
	    numerge++;
	}
	else {
	    input = load_file(p, input, 1);
	    if (!input) return(1);
	}
    }

    /* Add the last key = sort whole paragraph */
    ekey.text = "";
    ekey.neg = neg;
    ekey.circf = num;
    ekey.word = 0;
    if (last) last->previous = &ekey;
    else sort_keys = &ekey;

    /* MERGE option: do it */
    if (merge) {
	if (numerge == 0)
	    merge[numerge++].file = stdin;
        if (ofile) {
	    of = fopen(ofile, "w");
	    if (!of) {
	        fprintf(stderr, "****%s: ", progname);
	        perror(ofile);
	        return(1);
	    }
        }
        else of = stdout;
	return(merge_files(merge, numerge, of));
    }
    /* Verify files specified, defaut = stdin */
    else if (!input) {			/* No input file... */
	input = load_stream(0);
	if (!input) return(1);
    }

    /* Count the number of paragraphs */
    count=1, p=input;
    while (*p) {
	for (i=0; delim[i] && (p[i] == delim[i]); i++) ;
	if (delim[i] == 0) { 	/* New paragraph */
	    count++;
	    p += i;
	}
	p += strloc(p, '\n');
	if (*p) p++;
    }
    pars = malloc((count+1)*sizeof(char *));	/* Mod. V5.44 */

    /* Allocate the pointers to the paragraphs */
    p = input;
    a = pars;
    if (delim[0]) *(a++) = input;
    while (*p) {
	for (i=0; delim[i] && (p[i] == delim[i]); i++) ;
	if (delim[i] == 0) {	/* New paragraph */
	    if (p>input) p[-1] = 0;
	    while (--i>=0) *(p++) = 0;
	    /* V5.4: empty lines are ignored */
	    while (*p == '\n') *(p++) = 0;
	    *(a++) = p;
	}
	p += strloc(p, '\n');
	if (*p) p++;
    }
    *a = (char *)0;
    /* Remove ending newlines */
    while (p>input) {
	if (*--p == '\n') *p = 0;
	else break;
    }

    /* Remove the empty lines -- except first one... */
    for (a=b=pars+1; *a; a++) {
	p = *a;
	if (! *p) continue;
	*(b++) = *a;
    }
    *b = (char *)0;
    count = b-pars;

    /* Sort the file */
    qsort(pars+1, count-1, sizeof(char *), (SORT_FCT)pardif);

    /* Write out the file */
    if (ofile) {
	of = fopen(ofile, "w");
	if (!of) {
	    fprintf(stderr, "****%s: ", progname);
	    perror(ofile);
	    return(1);
	}
    }
    else of = stdout;
    for (a=pars; *a; a++) {
	fprintf(of, "%s%s\n", *a, delim);
	if (uniq && (a != pars)) while (a[1]) {
	    i = IGNORECASE ?
	        stucmp(a[0], a[1]) :
	        strcmp(a[0], a[1]);
	    if (i) break;
	    i++;
	}
    }

    return(0);
}

int parmerge(int argc, char **argv, int arg1)
/*++++++++++++++++
.PURPOSE  Main
.RETURNS  Status
.REMARKS  xgrep=2 for egrep
-----------------*/
{
    fprintf(stderr, "*****Program not yet accessible: %s\n",
      progname);
    return(1);
}


/* The various programs */
static struct { char *name; int (*prog)(); int arg1; } programs[] = {
    {/* [0] */  "parfile",   0 , 0},
    {/* [1] */  "pargrep",   pargrep, GREP },
    {/* [2] */  "paregrep",  pargrep, EGREP },
    {/* [3] */  "parfgrep",  pargrep, FGREP },
    {/* [4] */  "parindex",  parindex, 0 },
    {/* [5] */  "parsort",   parsort, 0 },
    {/* [6] */  "parmerge",  parmerge, 0 },
    {(char *)0, 0 }
};
static int parprog = 0;

/*==================================================================*/

int main(int argc, char **argv)
/*++++++++++++++++
.PURPOSE  Main
.RETURNS  to OS
.REMARKS
-----------------*/
{
  char *p;
  char *line;	/* Input */
  INT4 val[2];
  int  i;
  LINK1 *link;
  FILE *file ;

    /* Necessary Initialisation */
    word_fct = match_ascword ;		/* Default input = latex */

    /* V5.1: several programs */
    p = *argv;
    i = strlen(p);
    while ((--i>=0) && (p[i] != '/') && (p[i] != '\\')) ; ++i;
    for (parprog=0; programs[parprog].name; parprog++) {
	if (strcmp(programs[parprog].name, p+i) == 0) break;
    }
    if (!programs[parprog].name) {
	fprintf(stderr, "****Program not accessible: %s\n", *argv);
	exit(1);
    }

    progname = programs[parprog].name;

    if (programs[parprog].prog) {
        addIntro = 2;
	i = (*programs[parprog].prog)(argc, argv, programs[parprog].arg1);
	exit(i);
    }

    /* Take care of Options */
    while (--argc > 0) {
	p = *++argv;
	if (*p == '-') switch(p[1]) {	/* Option 	*/
	    case 'b':
		obyte = 1;
		if (grep) fprintf(stderr,
		  "+++ parfile: option '%s' ignored (incompatible with -e)\n",
		  p);
		continue;
	    case 'd':			/* Delimiter	*/
		delim = p+2;
		trc(delim, delim);
		continue;
	    case 'e':
		if (p[2] == 'c') {	/* Echo the Input Lines */
		    echo = "\n#";
		    continue;
		}
		if (p[2]) {		/* Regular expression   */
		    if (!ispunct(p[2])) goto bad_option;
		    p += 2 ;
		}
		else { argc--; p = *++argv ; }
		link = compile_expr(p, GREPdelimited) ;
		if (!link) exit(1);
		link->previous = grep;
		grep = link;
		if (og | obyte) fprintf(stderr,
		    "+++ parfile: -e option removes the -g/-g/-gs options\n") ;
		continue ;
	    case 'f':			/* Synonyms	*/
		++argv; --argc;
		fload(*argv);
		continue;
	    case 'g':			/* General indx	*/
		og = 1;
		if (grep) fprintf(stderr,
		  "+++ parfile: option '%s' ignored (incompatible with -e)\n",
		  p);
		if (strchr(p, 's')) og |= SEARCH_SILENT;
		if (strchr(p, 'b')) og |= SEARCH_BEST;
		while(isalpha(*p)) p++;
		if (isdigit(*p)) oglist = atoi(p);
		if (oglist == 0) /* Display ALL maching */
		    oglist = 99;
		continue;
	    case 'h':
	    	if (strcmp(p, "-help") == 0) { Help(); exit(0); }
		goto bad_option;
	    case 'i':			/* All indexes	*/
	    	if (isdigit(p[2]))  {
		    opti2 = p[2] - '0';
		    continue;
		}
	    	if (strcmp(p, "-index") == 0) { oc |= 1; continue; }
	    	if (strcmp(p, "-i0") == 0)    { oc |= 1; continue; }
	    	if (strcmp(p, "-inum") == 0)  { inumop |= 1; continue; }
	    	if (strcmp(p, "-itex") == 0)  { word_fct = match_texword ;
						itexop |= 1; continue; }
	    	if ((strncmp(p, "-ii-", 4) == 0)
	    	  ||(strncmp(p, "-i-", 3)  == 0)) {
		    oc |= 2;	/* Create General Index */
		    if (p[2] == 'i') { 		/* Ignore Case */
			optii = 1;
			oc |= 4;	/* For creation, too.. */
		    }
		    p++ ; p += strloc(p, '-') + 1 ;
		    oc1 = '%';
		    if (*p == '%') ogk = ++p;
		    continue;
	    	}
		if (p[2] == '/') {	/* RexExp	*/
		    oc |= 2;	/* Create General Index */
		    p += 2;
		    i = strlen(p); lreg = MAX(lreg, i);
		    link = compile_expr(p, GREPdelimited);
		    if (!link) exit(1);
		    link->previous = reg_keep;
		    reg_keep = link;
		    continue ;
	    	}
		if ((p[2] == '!') || (p[2] == '^')) {	/* ! RexExp	*/
		    oc |= 2;	/* Create General Index */
		    p += 3;
		    i = strlen(p); lreg = MAX(lreg, i);
		    link = compile_expr(p, GREPdelimited);
		    if (!link) exit(1);
		    link->previous = reg_leave;
		    reg_leave = link;
		    continue ;
	    	}
		goto bad_option ;
	    case 'l':			/* -lw    opt.	*/
		line = *++argv; --argc;
	        maxwordlen = atoi(line) ;
	        if ((maxwordlen < 2) || (maxwordlen > MAXWORDLEN)) {
		    fprintf(stderr, "*** parfile: %d is not a valid number ",
			maxwordlen) ;
		    fprintf(stderr, "for -lw, reset to %d\n", MAXWORDLEN) ;
		    maxwordlen = MAXWORDLEN ;
		}
		continue;
	    case 'm':			/* -merge opt.	*/
		mfile = *++argv; argc-- ;
		continue;
	    case 'n':		/* Show byte offset	*/
		o_n = p+2;
		continue;
	    case 'p':			/* -plain 	*/
		addIntro = atoi(p+2);
		while (isdigit(p[2])) p++;
		continue;
	    case 'r':			/* -replace	*/
		orm = 1;
		mfile = *++argv; argc-- ;
		continue;
	    case 's':		/* Option size		*/
		line = *++argv; --argc;
		i = atoi(line) ;
		switch(p[2]) {
		  case 'b':
		    sdata_size = check_range(i, 128, 32768, p) ;
		    continue ;
		  case 'p':
		    spart_size = check_range(i, 16, 1024, p) ;
		    continue ;
		  default:
		    goto bad_option ;
		}
	    case 't':		/* Option TeX output 	*/
		otex = 1 ;
		continue ;
	    case 'v': 			/* Verbose	*/
		if (strncmp(p, "-ver", 4) == 0) {
		    printf("%s (CDS) Version %s\n", progname, VERSION);
		    return(0);
		}
		if (isdigit(p[2])) ov = atoi(p+2) ;
		else ov = 10;		/* Flag > 10%	*/
		continue;
	    case 'w':		/* Weight Output	*/
	        o_w = p+2;
		continue;
	    case 'x':			/* Extension	*/
		if (p[2] == '.') xtension = p+2;
		else --argc, xtension = *++argv;
		continue;
	    default:
	    bad_option:
	    	fprintf(stderr, "*** parfile: Invalid Option: %s\n", p);
	    	exit(Usage());
	}
	if (grep) {			/* Regular Expression Search */
	    file = fopen(p, "r") ;
	    if (!file) {
		fprintf(stderr, "*** parfile" ) ;
		perror(p) ;
		exit(1) ;
	    }
	    grepPar(file, grep, 4) ;
	    fclose(file) ;
	    nfgrep++;
	    continue ;
	}
	if (parfiles >= (MAXFILES-1)) {
	    fprintf(stderr, "*** parfile: Too many (%d) parfiles: %s\n",
	    	parfiles, p);
	    exit(Usage());
	}
	parfile[parfiles].name = p;
	parfiles++;
    }

    /* WithRegExpr Search, it's finished -- unless stdin... */
    if (grep) {
	if (nfgrep == 0) grepPar(stdin, grep, 4) ;
	exit(0) ;
    }

	/* Check everything is here */
    if (!parfiles) {
	fprintf(stderr, "*** parfile: Missing parfile(s)\n");
	exit(Usage());
    }
    multiple = (parfiles > 1);
    for (p=delim; *p; p++) nnl += *p == '\n';

	/* Create Option */
    if (oc) {
   	for (i=0; i < parfiles; i++) {
   	    if (oc&1)
		if (CreateLindex(&parfile[i])) exit(1) ;
   	    if (oc&2)
		if (CreateGindex(&parfile[i])) exit(1) ;
   	}
  	exit(0);
    }

	/* Merge option	*/
    if (mfile)		exit(Merge(mfile));

	/* Retrieving data */
    /* if (ov && (og & SEARCH_SILENT)) { og &= ~SEARCH_SILENT; og |= 1;} */
    val[0] = 0;
    if (OpenParfile(&parfile)<0) exit(1);
    while ((line = Gets())) switch(line[0]){
      case '#':		/* Comment, just ignore */
	if (echo) if (echo) printf("%s%s\n", echo, line);
	continue;
      case '.':		/* General Key		*/
	if (echo) if (echo) printf("%s%s\n", echo, line);
	DisplayKey(line+1);
	continue;
      default:
	if (og) {
	    if (echo) if (echo) printf("%s%s\n", echo, line);
	    DisplayKey(line) ;
	    continue;
	}
			/* Get Line/Offset #	*/
	/* p = strstr(line, "%%"); */
	if (getNumbers(line, &val[1^multiple], 1+multiple))
	    Display(val[0], val[1]);
	/* else if (p && *p)
	    printf("%s \t(not found)", p) ;
	*/
	else continue;
	putchar('\n') ;
    }
    exit(0);
}
